# -*- coding: utf-8 -*-
from PyQt4.Qt import *
from Kittens.widgets import PYSIGNAL

import Timba
from MeqGUI import Grid, GUI, Plugins
from Timba.dmi import *
from Timba.utils import *
from MeqGUI.GUI.pixmaps import pixmaps
from Timba.TDL import TDLOptions

import configparser

import os.path

_dbg = verbosity(0,name='tdloptgui');
dprint = _dbg.dprint;
dprintf = _dbg.dprintf;

PROFILE_FILE = "tdlconf.profiles";

ColumnName = 0;
ColumnToolTip =1;
ColumnValue = 2;

class TDLOptionsDialog (QDialog,PersistentCurrier):
  """implements a floating window for TDL options""";
  def __init__ (self,parent,ok_label=None,ok_icon=None):
    QDialog.__init__(self,parent);
    self.setSizeGripEnabled(True);

    lo_main = QVBoxLayout(self);
    # add option treeWidget
    self._tw = QTreeWidget(self);
    self._tw.setColumnCount(3);
    self._tw.setRootIsDecorated(True);
    self._tw.setSortingEnabled(False);
    self._tw.setSelectionMode(QAbstractItemView.NoSelection);

    self._tw.header().setResizeMode(ColumnName,QHeaderView.ResizeToContents);
    self._tw.header().setResizeMode(ColumnToolTip,QHeaderView.ResizeToContents);
    self._tw.header().setResizeMode(ColumnValue,QHeaderView.Stretch);
    self._tw.header().resizeSection(ColumnValue,300);
    self._tw.header().hide();
    QObject.connect(self._tw.header(),SIGNAL("sectionResized(int,int,int)"),self._resize_dialog);
    self._allow_resize_dialog = True;
    # self._tw.setSizePolicy(QSizePolicy.MinimumExpanding,QSizePolicy.MinimumExpanding);
    # add signals
    QObject.connect(self._tw,SIGNAL("itemClicked(QTreeWidgetItem*,int)"),self._process_treewidget_click);
    QObject.connect(self._tw,SIGNAL("itemActivated(QTreeWidgetItem*,int)"),self._process_treewidget_press);
    # this is emitted when a job button is puhysed in the treewidget -- hide ourselves when this happens
    QObject.connect(self._tw,PYSIGNAL("closeWindow()"),self.hide);
    # self._tw.setMinimumSize(QSize(200,100));
    lo_main.addWidget(self._tw);
    # set geometry
    # self.setMinimumSize(QSize(200,100));
    # self.setGeometry(0,0,200,60);

    # add buttons on bottom
    lo_btn  = QHBoxLayout(None);
    if ok_label:
      if ok_icon:
        self.wok = QPushButton(ok_icon.icon(),ok_label,self);
      else:
        self.wok = QPushButton(ok_label,self);
      lo_btn.addWidget(self.wok);
      QObject.connect(self.wok,SIGNAL("clicked()"),self.accept);
    else:
      self.wok = None;
    # addspacer
    spacer = QSpacerItem(40,20,QSizePolicy.Expanding,QSizePolicy.Minimum);
    lo_btn.addItem(spacer)
    # add load button
    self.wload = QToolButton(self);
    self.wload.setText("Load");
    self.wload.setIcon(pixmaps.fileopen.icon());
    self.wload.setToolTip("""<P>Loads current option settings from a named <i>profile</I>.
          Profiles are kept in a file called %s in your current directory.</P>"""%PROFILE_FILE);
    self.wload.setPopupMode(QToolButton.InstantPopup);
    self.wload.setToolButtonStyle(Qt.ToolButtonTextBesideIcon);
    self.load_menu = QMenu(self);
    self.wload.setMenu(self.load_menu);
    lo_btn.addWidget(self.wload);
    # add save button
    self.wsave = QToolButton(self);
    self.wsave.setText("Save");
    self.wsave.setIcon(pixmaps.filesave.icon());
    self.wsave.setToolTip("""<P>Saves current option settings to a named <i>profile</I>.
          Profiles are stored in a file called %s in your current directory.</P>"""%PROFILE_FILE);
    self.wsave.setPopupMode(QToolButton.InstantPopup);
    self.wsave.setToolButtonStyle(Qt.ToolButtonTextBesideIcon);
    self.save_menu = QMenu(self);
    self.wsave.setMenu(self.save_menu);
    lo_btn.addWidget(self.wsave);
    # create the New profile action for the save menu
    self._qa_newprof = QAction("New profile...",self);
    QObject.connect(self._qa_newprof,SIGNAL("triggered()"),self._save_new_profile);
    # load profile configuration
    self.refresh_profiles();
    QObject.connect(self,PYSIGNAL("refreshProfiles()"),self.refresh_profiles);
    # add cancel button
    tb = QPushButton(pixmaps.red_round_cross.icon(),"Cancel",self);
    QObject.connect(tb,SIGNAL("clicked()"),self.hide);
    spacer = QSpacerItem(40,20,QSizePolicy.Expanding,QSizePolicy.Minimum);
    lo_btn.addItem(spacer)
    lo_btn.addWidget(tb);
    lo_main.addLayout(lo_btn);
    self.resize(QSize(600,480).expandedTo(self.minimumSizeHint()))

  def enableOkButton (self,enable):
    if self.wok:
      self.wok.setEnabled(enable);

  def adjustSizes (self):
    width = 300;  # minimal size of content section
    # add width of other sections
    for col in [0,1]:
      width += self._tw.header().sectionSize(col);
    # print width;
    self.resize(max(width,self.width()),self.height());

  def refresh_profiles (self):
    """Reloads the profiles file and repopulates load and save menus accordingly""";
    self.load_menu.clear();
    self.save_menu.clear();
    self.save_menu.addAction(self._qa_newprof);
    self.save_menu.addSeparator();
    self.profiles = configparser.RawConfigParser();
    try:
      self.profiles.read([PROFILE_FILE]);
    except:
      pass;
    self.wload.setEnabled(bool(self.profiles.sections()));
    for name in sorted(self.profiles.sections()):
      self.load_menu.addAction(name,self.curry(self._load_profile,name));
      self.save_menu.addAction(name,self.curry(self._save_profile,name));

  def _load_profile (self,name):
    if not self.profiles.has_section(name):
      QMessageBox.warning(self,"Error loading profile","""<P>Profile '%s' not found,
              check your %s file.</P>"""%(name,PROFILE_FILE));
      return;
    for name,value in self.profiles.items(name):
      try:
        TDLOptions.set_option(name,value,save=False,from_str=True);
      except ValueError:
        dprintf(0,"can't set '%s' to %s, ignoring",name,value);
    TDLOptions.save_config();

  def _save_profile (self,name):
    if self.profiles.has_section(name):
      if QMessageBox.warning(self,"Overwriting profile","""<P>Do you really want to
                          overwrite the existing profile '%s'?</P>"""%name,
                          QMessageBox.Ok|QMessageBox.Cancel,QMessageBox.Ok) != QMessageBox.Ok:
        return;
      newprof = False;
    else:
      self.profiles.add_section(name);
      newprof = True;
    TDLOptions.save_to_config(self.profiles,name);
    try:
      ff = open(PROFILE_FILE,"wt");
      for section in sorted(self.profiles.sections()):
        ff.write("[%s]\n"%section);
        for opt,value in sorted(self.profiles.items(section)):
          ff.write("%s = %s\n"%(opt.lower(),value));
        ff.write("\n");
      ff.close();
    except:
      dprintf(0,"error writing %s"%PROFILE_FILE);
      traceback.print_exc();
      QMessageBox.warning(self,"Error saving profile","""<P>There was an error writing to the %s file,
                           profile was not saved.</P>"""%PROFILE_FILE);
    if newprof:
      self.emit(PYSIGNAL("refreshProfiles()"));

  def _save_new_profile (self):
    default_name = os.path.splitext(os.path.basename(TDLOptions.current_scriptname))[0] + ":";
    name,ok = QInputDialog.getText(self,"Saving profile","Enter profile name",QLineEdit.Normal,default_name);
    if ok:
      self._save_profile(str(name));

  def _resize_dialog (self,section,*dum):
    if self._allow_resize_dialog and section<ColumnValue:
      self.adjustSizes();

  def accept (self):
    self.emit(PYSIGNAL("accepted()"));
    self.hide();

#  def show (self):
    #self._tw.adjustColumn(0);
    #self._tw.adjustColumn(1);
#    return QDialog.show(self);

#  def exec_ (self):_load_prof
#    #self._tw.adjustColumn(0);
#    #self._tw.adjustColumn(1);
#    return QDialog.exec_(self);

  def treeWidget (self):
    return self._tw;

  def clear (self):
    self._tw.clear();

  def addAction (self,label,callback,icon=None):
    # find last item
    last = self._tw.topLevelItem(self._tw.topLevelItemCount()-1);
    item = QTreeWidgetItem(self._tw,last);
    button = QPushButton(self._tw);
    button.setText(label);
    if icon:
      button.setIcon(icon.icon());
    self._tw.setItemWidget(item,0,button);
    button._on_click = callback;
    QObject.connect(button,SIGNAL("clicked()"),button._on_click);
    QObject.connect(button,SIGNAL("clicked()"),self.hide);

  def _process_treewidget_click (self,item,col,*dum):
    """helper function to process a click on a treeWidget item. Meant to be connected
    to the clicked() signal of a QListView""";
    if col == ColumnToolTip:
      tip = str(item.toolTip(ColumnToolTip));
      if tip:
        pos = self._tw.viewport().mapToGlobal(QPoint(self._tw.columnViewportPosition(ColumnToolTip),self._tw.visualItemRect(item).top()));
        QToolTip.showText(pos,tip,self._tw);
    else:
      getattr(item,'_on_click',lambda:None)();
      if getattr(item,'_close_on_click',False):
        self.hide();

  def _process_treewidget_press (self,item,col,*dum):
    """helper function to process a press on a treeWidget item. Meant to be connected
    to the pressed() signal of a QListView""";
    if col == ColumnToolTip:
      tip = str(item.toolTip(ColumnToolTip));
      if tip:
        pos = self._tw.viewport().mapToGlobal(QPoint(self._tw.columnViewportPosition(ColumnToolTip),self._tw.visualItemRect(item).top()));
        QToolTip.showText(pos,tip,self._tw);
    else:
      getattr(item,'_on_press',lambda:None)();




#!/usr/bin/env python3

# modules that are imported

#% $Id$ 

#
# Copyright (C) 2002-2007
# ASTRON (Netherlands Foundation for Research in Astronomy)
# and The MeqTree Foundation
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands, seg@astron.nl
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#  (c) 2013.				 (c) 2011.
#  National Research Council		 Conseil national de recherches
#  Ottawa, Canada, K1A 0R6 		 Ottawa, Canada, K1A 0R6
#
#  This software is free software;	 Ce logiciel est libre, vous
#  you can redistribute it and/or	 pouvez le redistribuer et/ou le
#  modify it under the terms of	         modifier selon les termes de la
#  the GNU General Public License	 Licence Publique Generale GNU
#  as published by the Free		 publiee par la Free Software
#  Software Foundation; either	 	 Foundation (version 3 ou bien
#  version 2 of the License, or	 	 toute autre version ulterieure
#  (at your option) any later	 	 choisie par vous).
#  version.
#
#  This software is distributed in	 Ce logiciel est distribue car
#  the hope that it will be		 potentiellement utile, mais
#  useful, but WITHOUT ANY		 SANS AUCUNE GARANTIE, ni
#  WARRANTY; without even the	 	 explicite ni implicite, y
#  implied warranty of			 compris les garanties de
#  MERCHANTABILITY or FITNESS FOR	 commercialisation ou
#  A PARTICULAR PURPOSE.  See the	 d'adaptation dans un but
#  GNU General Public License for	 specifique. Reportez-vous a la
#  more details.			 Licence Publique Generale GNU
#  					 pour plus de details.
#
#  You should have received a copy	 Vous devez avoir recu une copie
#  of the GNU General Public		 de la Licence Publique Generale
#  License along with this		 GNU en meme temps que ce
#  software; if not, contact the	 logiciel ; si ce n'est pas le
#  Free Software Foundation, Inc.	 cas, communiquez avec la Free
#  at http://www.fsf.org.		 Software Foundation, Inc. au
#						 http://www.fsf.org.
#
#  email:				 courriel:
#  business@hia-iha.nrc-cnrc.gc.ca	 business@hia-iha.nrc-cnrc.gc.ca
#
#  National Research Council		 Conseil national de recherches
#      Canada				    Canada
#  Herzberg Institute of Astrophysics	 Institut Herzberg d'astrophysique
#  5071 West Saanich Rd.		 5071 West Saanich Rd.
#  Victoria BC V9E 2E7			 Victoria BC V9E 2E7
#  CANADA					 CANADA
#
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import numpy

from qwt.qt.QtGui import QApplication, QGridLayout, QLabel, QWidget, QFont
from qwt.qt.QtCore import Qt

HAS_TIMBA = False
try:
  from Timba import utils
  from Timba.Meq import meqds
  from Timba.Meq.meqds import mqs
  from MeqGUI.GUI.pixmaps import pixmaps
  from MeqGUI.GUI import widgets
  from MeqGUI.GUI.browsers import *
  from MeqGUI import Grid

  from MeqGUI.Plugins.display_image_qt5 import QwtImageDisplay
  from MeqGUI.Plugins.QwtPlotImage_qt5 import QwtPlotImage
  from MeqGUI.Plugins.QwtColorBar_qt5 import QwtColorBar
  from MeqGUI.Plugins.ND_Controller_qt5 import ND_Controller
  import MeqGUI.Plugins.plotting_functions_qt5 as plot_func
  #from MeqGUI.Plugins.plot_printer_qt5 import *


  from Timba.utils import verbosity
  _dbg = verbosity(0,name='array_plotter');
  _dprint = _dbg.dprint;
  _dprintf = _dbg.dprintf;
  HAS_TIMBA = True
except:
  pass

global has_vtk
has_vtk = False

class ArrayPlotter(GriddedPlugin):
  """ a class to plot raw arrays contained within a Meq tree """

  _icon = pixmaps.bars3d
  viewer_name = "Array Plotter";
  def is_viewable (data):
    return len(data) > 0;
  is_viewable = staticmethod(is_viewable);

  def __init__(self,gw,dataitem,cellspec={},**opts):
    GriddedPlugin.__init__(self,gw,dataitem,cellspec=cellspec);
    if HAS_TIMBA: _dprint(3,'** array_plotter: starting init')
    self.layout_parent = None
    self.layout_created = False
    self.twoD_plotter = None
    self.status_label = None
    self.ND_plotter = None
    self.ND_Controls = None
    self.png_number = 0

    if dataitem and dataitem.data is not None:
      self.set_data(dataitem);

#  def __del__(self):
#    print "in destructor"

  def create_2D_plotter(self):
    if not self.ND_plotter is None:
      self.ND_plotter.close()
      self.ND_plotter = None
    self.twoD_plotter, self.plotPrinter = plot_func.create_2D_Plotters(self.layout, self.layout_parent)
    self.twoD_plotter.colorbar_needed.connect(self.set_ColorBar)
    self.twoD_plotter.show_ND_Controller.connect(self.ND_controller_showDisplay)
    self.twoD_plotter.show_3D_Display.connect(self.show_3D_Display)
    self.twoD_plotter.do_print.connect(self.plotPrinter.do_print)
    self.twoD_plotter.save_display.connect(self.grab_display)
    # create status label display
    self.status_label = Qt.QLabel(self.layout_parent)
    sansFont = QFont( "Helvetica [Cronyx]", 8 )
    self.status_label.setFont(sansFont)
#   self.layout.addMultiCellWidget(self.status_label,1,1,0,2)
    self.layout.addWidget(self.status_label,1,0,1,2)
    self.status_label.setText("Move the mouse within the plot canvas"
                            " to show the cursor position.")
    self.status_label.show()
    self.twoD_plotter.status_update.connect(self.update_status)
  # create_2D_plotter

  def update_status(self, status):
     if not status is None:
       self.status_label.setText(str(status))

  def grab_display(self, title):
    self.png_number = self.png_number + 1
    png_str = str(self.png_number)
    if title is None:
      save_file = './meqbrowser' + png_str + '.png'
    else:
      save_file = title + png_str + '.png'
    save_file_no_space= save_file.replace(' ','_')
    result = Qt.QPixmap.grabWidget(self.layout_parent).save(save_file_no_space, "PNG")

  def create_layout_stuff(self):
    """ create grid layouts into which plotter widgets are inserted """
    if self.layout_parent is None or not self.layout_created:
      self.layout_parent = Qt.QWidget(self.wparent())
      self.layout = Qt.QGridLayout(self.layout_parent)
      self.set_widgets(self.layout_parent,self.dataitem.caption,icon=self.icon())
      self.layout_created = True
    self._wtop = self.layout_parent;

  def wtop (self):
    return self._wtop

  def set_data (self,dataitem,default_open=None,**opts):
    """ this function is the callback interface to the meqbrowser and
        handles new incoming data """

    self.dataitem = dataitem
# first create layout stuff
    self.create_layout_stuff()

# now create plotter and colorbar
    if self.twoD_plotter is None:
      self.create_2D_plotter()
    self.plot_2D_array()

  def test_scalar_value (self, data_array, data_label):
    """ test if incoming 'array' contains only a scalar value """
# do we have a scalar?
    is_scalar = False
    scalar_data = 0.0
    try:
      shape = data_array.shape
      if HAS_TIMBA: _dprint(3,'data_array shape is ', shape)
    except:
      is_scalar = True
      scalar_data = data_array
    if not is_scalar and len(shape) == 1:
      if shape[0] == 1:
        is_scalar = True
        scalar_data = data_array[0]
    if not is_scalar:
      test_scalar = True
      for i in range(len(shape)):
        if shape[i] > 1:
          test_scalar = False
      is_scalar = test_scalar
      if is_scalar:
        scalar_data = data_array
    if is_scalar:
      self.twoD_plotter.report_scalar_value(data_label, scalar_data)
      return True
    else:
      return False
# enable & highlight the cell
    self.enable();
    self.flash_refresh();

  def plot_2D_array (self):

    self.data = self.dataitem.data

# do we have a scalar?
    if self.test_scalar_value(self.data, 'incoming data'):
      return 

# first figure out the actual rank of the array we are plotting

    self.actual_rank = 0
    self.array_shape = self.dataitem.data.shape
    self.array_rank = self.dataitem.data.ndim
    for i in range(len(self.array_shape)):
      if self.array_shape[i] > 1:
        self.actual_rank = self.actual_rank + 1
    self.array_selector = None
    if self.actual_rank > 2:
      self.twoD_plotter.set_original_array_rank(self.actual_rank)
      self.set_ND_controls()

# reset colorbar just in case
    if self.actual_rank >= 2:
      self.twoD_plotter.reset_color_bar(True)

# pass array to the plotter
    if self.array_rank > 2:
      if self.array_selector is None:
        second_axis = None
        first_axis = None
        for i in range(self.array_rank-1,-1,-1):
          if self.data.shape[i] > 1:
            if second_axis is None:
              second_axis = i
            else:
              if first_axis is None:
                first_axis = i
        if not second_axis is None:
          self.array_selector = plot_func.create_array_selector(None, self.array_rank, self.data.shape, first_axis,second_axis, -1)
      self.array_tuple = tuple(self.array_selector)
      self.twoD_plotter.array_plot(self.data[self.array_tuple],data_label='data')
    else:
      self.twoD_plotter.array_plot(self.data,data_label='data')


  def plot_3D_array (self, display_flag_3D):
# first figure out the actual rank of the array we are plotting
    self.actual_rank = 0
    self.array_shape = self.dataitem.data.shape
    self.array_rank = self.dataitem.data.ndim
    for i in range(len(self.array_shape)):
      if self.array_shape[i] > 1:
        self.actual_rank = self.actual_rank + 1
    if display_flag_3D and self.actual_rank > 2:
      self.set_ND_controls(None, None, num_axes=3)
      self.array_selector = None
    else:
      if not self.ND_Controls is None:
        self.ND_Controls.setParent(Qt.QWidget())
        self.ND_Controls = None
# pass initial array to the plotter
    if self.array_rank > 2:
      self.data = self.dataitem.data
      if self.array_selector is None:
        third_axis = None
        second_axis = None
        first_axis = None
        for i in range(self.array_rank-1,-1,-1):
          if self.data.shape[i] > 1:
            if third_axis is None:
              third_axis = i
            elif second_axis is None:
              second_axis = i
            else:
              if first_axis is None:
                first_axis = i
        if not first_axis is None and not second_axis is None and not third_axis is None:
          self.array_selector = plot_func.create_array_selector(None, self.array_rank, self.data.shape, first_axis,second_axis,third_axis)
      self.array_tuple = tuple(self.array_selector)
      plot_array =  self.data[self.array_tuple]
    else:
      plot_array = self.data
    if plot_array.min() == plot_array.max():
      return
    else:
      self.ND_plotter.array_plot(plot_array, data_label='data ')

# enable & highlight the cell
    self.enable();
    self.flash_refresh();

  def setSelectedAxes (self,first_axis, second_axis, third_axis=-1):
    """ update the selected axes of an N-dimensional array
        and display the selected sub-array.
    """
    self.array_selector = plot_func.create_array_selector(self.twoD_plotter, self.array_rank, self.array_shape, first_axis,second_axis,third_axis)
    self.array_tuple = tuple(self.array_selector)
    if not self.twoD_plotter is None:
      # reset various parameters and fields
      self.twoD_plotter.reset_zoom()
      self.twoD_plotter.delete_cross_sections()
      self.twoD_plotter.reset_color_bar(True)
      self.twoD_plotter.array_plot(self.data[self.array_tuple],data_label='data')
    else:
      if not self.ND_plotter is None:
        self.ND_plotter.delete_vtk_renderer()
        self.ND_plotter.array_plot(self.data[self.array_tuple],data_label='data ')
        axis_parms = []
        axis_parms.append('axis ' + str(first_axis))
        axis_parms.append('axis ' + str(second_axis))
        axis_parms.append('axis ' + str(third_axis))
        self.ND_plotter.setAxisParms(axis_parms)

  def setArraySelector (self,lcd_number, slider_value, display_string=None):
    """ update the subsection of an N-dimensional array 
        and display the selected subarray.
    """
    self.array_selector[lcd_number] = slider_value
    self.array_tuple = tuple(self.array_selector)
    if not self.twoD_plotter is None:
      # reset color bar request - just in case
      self.twoD_plotter.reset_color_bar(True)
      self.twoD_plotter.array_plot(self.data[self.array_tuple],data_label='data')
    else:
      self.ND_plotter.array_plot(self.data[self.array_tuple],data_label='data ')

  def ND_controller_showDisplay(self, show_self):
    """ tells ND_Controller to show or hide itself """
    if not self.ND_Controls is None:
      self.ND_Controls.showDisplay(show_self)

  def set_ND_controls (self, labels=None, parms=None, num_axes=2):
    """ this function adds the extra GUI control buttons etc if we are
        displaying data for a numpy array of dimension 3 or greater """

    self.ND_Controls = plot_func.create_ND_Controls(self.layout, self.layout_parent, self.array_shape, self.ND_Controls, self.ND_plotter, labels, parms, num_axes)

    self.ND_Controls.sliderValueChanged.connect(self.setArraySelector)
    self.ND_Controls.defineSelectedAxes.connect(self.setSelectedAxes)

  def show_3D_Display(self, display_flag_3D):
    if not has_vtk:
      return
    self.twoD_plotter = delete_2D_Plotters(self.colorbar, self.twoD_plotter)
    self.status_label.setParent(Qt.QWidget())
    self.status_label = None

    if self.ND_plotter is None:
      self.ND_plotter = plot_func.create_ND_Plotter (self.layout, self.layout_parent)
      self.ND_plotter.show_2D_Display.connect(self.show_2D_Display)
      self.ND_plotter.show_ND_Controller.connect(self.ND_controller_showDisplay)
    else:
      self.ND_plotter.delete_vtk_renderer()
      self.ND_plotter.show_vtk_controls()

# create 3-D Controller
    self.plot_3D_array(display_flag_3D)

  def show_2D_Display(self, display_flag):
    if HAS_TIMBA: _dprint(3, 'in show_2D_Display ')
    self.create_2D_plotter()
# create 3-D Controller appropriate for 2-D screen displays
    self.plot_2D_array()

  def set_ColorBar (self):
    """ this function adds a colorbar for 2-D displays """
    # create two color bars in case we are displaying complex arrays
    self.colorbar = plot_func.create_ColorBar(self.layout, self.layout_parent, self.twoD_plotter, self.plotPrinter)

Grid.Services.registerViewer(array_class,ArrayPlotter,priority=10)



#-*- coding: utf-8 -*-
# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2007 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from qt import SIGNAL, Qt, QColorDialog

import os
import qt
import re
import sys
import types
from .dialog_window_config import *
#from tabdialog import *


connect = qt.QObject.connect

# helper functions to WidgetSettingsDialog
_grid_styles = [Qt.SolidLine, Qt.DashLine, Qt.DotLine, Qt.DashDotLine,
                Qt.DashDotDotLine]

def _item_to_style(item):
    """Returns what line-style is on item in style combo box."""
    return _grid_styles[item]

def _style_to_item(style):
    """Returns which item in line-style combo box holds style."""
    return _grid_styles.index(style)


class WidgetSettingsDialog:

    def __init__(self, actual_parent,gui_parent):
        self._widget = dialog_window_config(parent=gui_parent, mother=self)

        self._parent_widget = actual_parent

        # labels settings
        plot_parms = self._parent_widget.getPlotParms()
        self._widget.child('line_edit_window_title').setText(plot_parms['window_title'])
        self._widget.child('line_edit_x_axis_label').setText(plot_parms['x_title'])
        self._widget.child('line_edit_y_axis_label').setText(plot_parms['y_title'])

        # grid settings
#       gridsettings = self._parent_widget.get_grid_settings()
        self._maj_colour_button = self._widget.child('push_button_maj_grid_colour')
        self._min_colour_button = self._widget.child('push_button_min_grid_colour')
        self._maj_x_enabled = self._widget.child('check_box_x_grid_maj_enable')
        self._maj_y_enabled = self._widget.child('check_box_y_grid_maj_enable')
        self._min_x_enabled = self._widget.child('check_box_x_grid_min_enable')
        self._min_y_enabled = self._widget.child('check_box_y_grid_min_enable')
#       self._maj_x_enabled.setChecked(gridsettings[0])
#       self._maj_y_enabled.setChecked(gridsettings[1])
#       self._maj_colour_button.setPaletteBackgroundColor(gridsettings[2])
#       self._widget.child('spin_box_maj_grid_width').setValue(gridsettings[3])
#       self._widget.child('combo_box_maj_grid_style').setCurrentItem(_style_to_item(gridsettings[4]))
#       self._min_x_enabled.setChecked(gridsettings[5])
#       self._min_y_enabled.setChecked(gridsettings[6])
#       self._min_colour_button.setPaletteBackgroundColor(gridsettings[7])
#       self._widget.child('spin_box_min_grid_width').setValue(gridsettings[8])
#       self._widget.child('combo_box_min_grid_style').setCurrentItem(_style_to_item(gridsettings[9]))

        # set correct initial state on gui elements for grid settings
        self._slot_enable_major_grid()
        self._slot_enable_minor_grid() 

        # axes settings
#       axessettings = self._parent_widget.get_axes_settings()
#       self._widget.child('check_box_x_auto_scale').setChecked(axessettings[0])
        self._x_min = self._widget.child('line_edit_x_min')
        self._x_max = self._widget.child('line_edit_x_max')
        self._y_min = self._widget.child('line_edit_y_min')
        self._y_max = self._widget.child('line_edit_y_max')
        self._ratio = self._widget.child('line_edit_ratio')
        self._x_min.setText(str(plot_parms['axis_xmin']))
        self._x_max.setText(str(plot_parms['axis_xmax']))
#       self._widget.child('check_box_y_auto_scale').setChecked(axessettings[3])
        self._y_min.setText(str(plot_parms['axis_ymin']))
        self._y_max.setText(str(plot_parms['axis_ymax']))
#       self._widget.child('check_box_ratio').setChecked(axessettings[6])
#       self._ratio.setText(str(axessettings[7]))

        self._connect_signals()
        self._widget.show()



    def _connect_signals(self):
        connect(self._widget.child('buttonOk'), SIGNAL('clicked()'),
                self._slot_ok_pressed)
        
        # grid connections
        connect(self._maj_colour_button, SIGNAL('clicked()'),
                self._slot_major_grid_colour)
        connect(self._min_colour_button, SIGNAL('clicked()'),
                self._slot_minor_grid_colour)
        connect(self._maj_x_enabled, SIGNAL('clicked()'),
                self._slot_enable_major_grid)
        connect(self._maj_y_enabled, SIGNAL('clicked()'),
                self._slot_enable_major_grid)
        connect(self._min_x_enabled, SIGNAL('clicked()'),
                self._slot_enable_minor_grid)
        connect(self._min_y_enabled, SIGNAL('clicked()'),
                self._slot_enable_minor_grid)

    def _slot_ok_pressed(self):
        settings = {}
        
        # labels settings
        settings['window_title'] = str(self._widget.child('line_edit_window_title').text())
        settings['x_title'] = str(self._widget.child('line_edit_x_axis_label').text())
        settings['y_title'] = str(self._widget.child('line_edit_y_axis_label').text())

        # grid settings
        settings['x_grid_maj_enable'] = str(int(self._maj_x_enabled.isChecked()))
        settings['x_grid_min_enable'] = str(int(self._min_x_enabled.isChecked()))
        settings['y_grid_maj_enable'] = str(int(self._maj_y_enabled.isChecked()))
        settings['y_grid_min_enable'] = str(int(self._min_y_enabled.isChecked()))
        settings['maj_grid_style'] = str(self._widget.child('combo_box_maj_grid_style').currentItem())
        settings['min_grid_style'] = str(self._widget.child('combo_box_min_grid_style').currentItem())
        settings['maj_grid_width'] = str(self._widget.child('spin_box_maj_grid_width').value())
        settings['min_grid_width'] = str(self._widget.child('spin_box_min_grid_width').value())
        settings['maj_grid_colour'] = str(self._maj_colour_button.paletteBackgroundColor().name())
        settings['min_grid_colour'] = str(self._min_colour_button.paletteBackgroundColor().name())

        # axes settings
        settings['x_auto_scale'] = str(int(self._widget.child('check_box_x_auto_scale').isChecked()))
        settings['y_auto_scale'] = str(int(self._widget.child('check_box_y_auto_scale').isChecked()))
        settings['axis_xmin'] = str(self._x_min.text())
        settings['axis_xmax'] = str(self._x_max.text())
        settings['axis_ymin'] = str(self._y_min.text())
        settings['axis_ymax'] = str(self._y_max.text())

        # apply settings
        self._parent_widget.setPlotParms(settings)

    
    def _set_colour_on_button(self, button):
        """Asks the user to choose a colour, then sets it on button."""
        colour = QColorDialog.getColor(button.paletteBackgroundColor(),
                                       self._widget)
        if colour.isValid():
            button.setPaletteBackgroundColor(colour)

    def _slot_major_grid_colour(self):
        self._set_colour_on_button(self._maj_colour_button)
    
    def _slot_minor_grid_colour(self):
        self._set_colour_on_button(self._min_colour_button)

    def _set_grid_widgets_enabled(self, mode, enable):
        for child in ['push_button_%s_grid_colour'%mode,
                      'spin_box_%s_grid_width'%mode,
                      'combo_box_%s_grid_style'%mode,
                      'text_label_%s_grid_colour'%mode,
                      'text_label_%s_grid_width'%mode,
                      'text_label_%s_grid_style'%mode]:
            self._widget.child(child).setEnabled(enable)

    def _slot_enable_major_grid(self):
        if self._maj_x_enabled.isChecked() or self._maj_y_enabled.isChecked():
            self._set_grid_widgets_enabled('maj', True)
        else:
            self._set_grid_widgets_enabled('maj', False)

    def _slot_enable_minor_grid(self):
        if self._min_x_enabled.isChecked() or self._min_y_enabled.isChecked():
            self._set_grid_widgets_enabled('min', True)
        else:
            self._set_grid_widgets_enabled('min', False)

    def _ensure_valid_float_strings(self, string):
        """Returns a string which is parseable by float() from the given string"""
        try:
            return str(float(re.match('[+|-]?[0-9]+[\.[0-9]*]?', string).group()))
        except:
            return '0.0'

    def _slot_x_max(self):
        x_max = str(self._x_max.text())
        self._x_max.setText(self._ensure_valid_float_strings(x_max))

    def _slot_x_min(self):
        x_min = str(self._x_min.text())
        self._x_min.setText(self._ensure_valid_float_strings(x_min))

    def _slot_y_max(self):
        y_max = str(self._y_max.text())
        self._y_max.setText(self._ensure_valid_float_strings(y_max))

    def _slot_y_min(self):
        y_min = str(self._y_min.text())
        self._y_min.setText(self._ensure_valid_float_strings(y_min))

    def _slot_ratio(self):
        ratio = str(self._ratio.text())
        self._ratio.setText(self._ensure_valid_float_strings(ratio))





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#
#% $Id: tdlgui.py 6822 2009-03-03 15:13:52Z oms $ 
#
#
# Copyright (C) 2002-2007
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import Timba
from MeqGUI import Grid, GUI, Plugins
from Timba.dmi import *
from Timba.utils import *
from MeqGUI import Grid
from MeqGUI.GUI import browsers
from MeqGUI.GUI.pixmaps import pixmaps
from MeqGUI.GUI import app_proxy_gui
from Timba.Meq import meqds
from Timba import TDL
import Timba.TDL.Settings
import Timba.TDL.Compile
from Timba.TDL import TDLOptions

from PyQt4.Qt import *
from Kittens.widgets import PYSIGNAL

import imp
import sys
import re
import traceback
import os
import os.path
import tempfile

_dbg = verbosity(0,name='tdlgui');
_dprint = _dbg.dprint;
_dprintf = _dbg.dprintf;



def _file_mod_time (path):
  try:
    return os.stat(path).st_mtime;
  except IOError:
    return None;

# flag: sync to external editor
_external_sync = True;
def set_external_sync (value):
  global _external_sync;
  _external_sync = value;

# this is information about ourselves
_MODULE_FILENAME = Timba.utils.extract_stack()[-1][0];
_MODULE_DIRNAME = os.path.dirname(_MODULE_FILENAME);

class TDLEditor (QFrame,PersistentCurrier):
  SupportsLineNumbers = True;
  SubErrorMarker = 0;
  ErrorMarker = 1;
  CurrentErrorMarker = 2;
  # a single editor always has the focus
  current_editor = None;

  def __init__ (self,parent,close_button=False,error_window=None):
    QFrame.__init__(self,parent);
    self._enabled = True;
    toplo = QVBoxLayout(self);
    self.setSizePolicy(QSizePolicy.Expanding,QSizePolicy.Expanding);
    splitter = QSplitter(Qt.Vertical,self);
    toplo.addWidget(splitter);
    splitter.setSizePolicy(QSizePolicy.Expanding,QSizePolicy.Expanding);
    splitter.setChildrenCollapsible(False);

    # figure out our app_gui parent
    self._appgui = app_proxy_gui.appgui(parent);

    # create an editor box
    editor_box = QFrame(splitter);
    lo = QVBoxLayout(editor_box);

    # find main window to associate our toolbar with
    self._toolbar = QToolBar("TDL tools",self._appgui,editor_box);
    lo.addWidget(self._toolbar);

    #### populate toolbar

    # Exec button and menu
    self._tb_jobs = QToolButton(self._toolbar);
    self._tb_jobs.setIconSet(pixmaps.gear.iconset());
    self._tb_jobs.setTextLabel("Exec");
    self._tb_jobs.setUsesTextLabel(True);
    self._tb_jobs.setTextPosition(QToolButton.BesideIcon);
    QToolTip.add(self._tb_jobs,"Access run-time options & jobs defined by this TDL script");
    self._tb_jobs.hide();
    
    jobs = self._jobmenu = TDLOptionsDialog(self);
    jobs.setCaption("TDL Jobs & Runtime Options");
    jobs.setIcon(pixmaps.gear.pm());
    jobs.hide();
    QObject.connect(self._tb_jobs,SIGNAL("clicked()"),jobs.exec_loop);

    # save menu and button
    self._tb_save = QToolButton(self._toolbar);
    self._tb_save.setIconSet(pixmaps.file_save.iconset());
    QToolTip.add(self._tb_save,"Saves script. Click on the down-arrow for other options.");
    savemenu = QPopupMenu(self);
    self._tb_save.setPopup(savemenu);
    self._tb_save.setPopupDelay(0);
    self._tb_save._modified_color = QColor("yellow");
    qa_save = QAction(pixmaps.file_save.iconset(),"&Save script",Qt.ALT+Qt.Key_S,self);
    QObject.connect(qa_save,SIGNAL("activated()"),self._save_file);
    QObject.connect(self._tb_save,SIGNAL("clicked()"),self._save_file);
    qa_save.addTo(savemenu);
    qa_save_as = QAction(pixmaps.file_save.iconset(),"Save script &as...",0,self);
    QObject.connect(qa_save_as,SIGNAL("activated()"),self.curry(self._save_file,save_as=True));
    qa_save_as.addTo(savemenu);
    qa_revert = self._qa_revert = QAction("Revert to saved",0,self);
    QObject.connect(qa_revert,SIGNAL("activated()"),self._revert_to_saved);
    qa_revert.addTo(savemenu);

    # run menu and button
    self._tb_run = QToolButton(self._toolbar);
    self._tb_run.setIconSet(pixmaps.blue_round_reload.iconset());

    self._tb_runmenu = QPopupMenu(self);
    self._tb_run.setPopup(self._tb_runmenu);
    self._tb_run.setPopupDelay(0);
    self._qa_runmain = QAction(pixmaps.blue_round_reload.iconset(),
                              "&Save & compile main script",Qt.ALT+Qt.Key_R,self);
    QObject.connect(self._qa_runmain,SIGNAL("activated()"),self._run_main_file);
    QObject.connect(self._tb_run,SIGNAL("clicked()"),self._import_main_file);
    self._qa_runmain.addTo(self._tb_runmenu);
    qa_runthis_as = QAction(pixmaps.blue_round_reload.iconset(),"Save & run this script as main script...",0,self);
    qa_runthis_as.setToolTip("Saves and recompiles this script as a top-level TDL script");
    QObject.connect(qa_runthis_as,SIGNAL("activated()"),self._import_as_main_file);
    qa_runthis_as.addTo(self._tb_runmenu);

    # Compile-time options and menu
    #self._tb_opts = QAction(pixmaps.wrench.iconset(),
                            #"Options",Qt.ALT+Qt.Key_O,self);
    #self._tb_opts.setToggleAction(True);
    #self._tb_opts.setToolTip("Access compile-time options for this TDL script");
    #self._tb_opts.addTo(self._toolbar);
    # Compile-time options and menu
    self._tb_opts = QToolButton(self._toolbar);
    self._tb_opts.setIconSet(pixmaps.wrench.iconset());
    self._tb_opts.setTextLabel("Options");
    self._tb_opts.setUsesTextLabel(True);
    self._tb_opts.setTextPosition(QToolButton.BesideIcon);
    QToolTip.add(self._tb_opts,"Access compile-time options for this TDL script");
    # self._tb_opts.hide();

    opts = self._options_menu = TDLOptionsDialog(parent,
            ok_label="Compile",ok_icon=pixmaps.blue_round_reload);
    opts.setCaption("TDL Compile-time Options");
    opts.setIcon(pixmaps.wrench.pm());
    QObject.connect(opts,PYSIGNAL("accepted()"),self._run_main_file);
    opts.hide();
    QObject.connect(self._tb_opts,SIGNAL("clicked()"),opts.show);
    
    self._qa_recompile = qa_recomp = QAction(pixmaps.blue_round_reload.iconset(),"Re&compile script to apply new options",0,self);
    qa_recomp.setToolTip("You must recompile this script for new options to take effect");
    QObject.connect(qa_recomp,SIGNAL("activated()"),self._run_main_file);

    self._toolbar.addSeparator();
    self._poslabel = QLabel(self._toolbar);
    width = self._poslabel.fontMetrics().width("L:999 C:999");
    self._poslabel.setMinimumWidth(width);
    self._pathlabel = QLabel(self._toolbar);
    self._pathlabel.setAlignment(Qt.AlignRight);
    self._pathlabel.setIndent(10);
    if close_button:
      if not isinstance(close_button,QIconSet):
        close_button = pixmaps.remove.iconset();
      self._qa_close = QAction(close_button,"&Close file",Qt.ALT+Qt.Key_W,self);
      QObject.connect(self._qa_close,SIGNAL("activated()"),self._file_closed);
      self._qa_close.addTo(self._toolbar);
    self._toolbar.setStretchableWidget(self._pathlabel);

    #### add editor window

    self._editor = QextScintilla(editor_box);
    lo.addWidget(self._editor);
    self._editor.setSizePolicy(QSizePolicy.Expanding,QSizePolicy.Expanding);
    self._lexer = QextScintillaLexerPython(self);
    self._editor.setLexer(self._lexer);
    self._editor_fontadjust = -6;
    self.adjust_editor_font();
    self._editor.markerDefine(QextScintilla.RightTriangle,self.ErrorMarker);
    self._editor.markerDefine(QextScintilla.RightTriangle,self.SubErrorMarker);
    self._editor.markerDefine(QextScintilla.RightTriangle,self.CurrentErrorMarker);
    self._editor.setMarkerForegroundColor(QColor("red"),self.ErrorMarker);
    self._editor.setMarkerForegroundColor(QColor("grey"),self.SubErrorMarker);
    self._editor.setMarkerForegroundColor(QColor("red"),self.CurrentErrorMarker);
    self._editor.setMarkerBackgroundColor(QColor("red"),self.CurrentErrorMarker);
    self._editor.setMarginSensitivity(1,True);
    QObject.connect(self._editor,SIGNAL("marginClicked(int,int,Qt::ButtonState)"),self._process_margin_click);
    QObject.connect(self._editor,SIGNAL("textChanged()"),self._text_changed);
    QObject.connect(self._editor,SIGNAL("modificationChanged(bool)"),self._text_modified);
    QObject.connect(self._editor,SIGNAL("cursorPositionChanged(int,int)"),self._display_cursor_position);
    # QObject.connect(self._editor,SIGNAL("textChanged()"),self._clear_transients);

    # add message window
    self._message_box = QFrame(editor_box);
    lo.addWidget(self._message_box);
    self._message_box.setSizePolicy(QSizePolicy.Expanding,QSizePolicy.Preferred);
    self._message_box.setFrameStyle(QFrame.Panel+QFrame.Sunken);
    self._message_box.setLineWidth(2);
    mblo = QVBoxLayout(self._message_box);
    msgb1 = QHBox(self._message_box);
    mblo.addWidget(msgb1);
    self._message_icon = QToolButton(msgb1);
    self._message = QLabel(msgb1);
    self._message.setTextFormat(Qt.RichText);
    self._message.setMargin(0);
    self._message.setSizePolicy(QSizePolicy.Expanding,QSizePolicy.Preferred);
    # self._message_icon.setAlignment(Qt.AlignTop);
    # self._message_icon.setMargin(4);
    self._message_icon.setAutoRaise(True);
    self._message_icon.setSizePolicy(QSizePolicy.Fixed,QSizePolicy.Fixed);
    QObject.connect(self._message_icon,SIGNAL("clicked()"),self.clear_message);
    QToolTip.add(self._message_icon,"Click here to clear the message");
    self._message_widgets = [];
    self._message_transient = False;

    # figure out if we already have an error box to attach to
    self._error_window = error_window or getattr(parent,'_tdlgui_error_window',None);
    if self._error_window:
      #self._resize_errwin = False;
      pass;
    else:
      # otherwise create an error floater
      self._error_window = TDLErrorFloat(parent);
      setattr(parent,'_tdlgui_error_window',self._error_window);
      # self._resize_errwin = True;
    QObject.connect(self._error_window,PYSIGNAL("hasErrors()"),self._reset_errors);
    QObject.connect(self._error_window,PYSIGNAL("showError()"),self.show_error);

    # set filename
    self._filename = None;       # "official" path of file (None if new script not yet saved)
    self._mainfile = None;       # if not None, then we're "slaved" to a main file (see below)
    self._file_disktime = None;  # modtime on disk when file was loaded
    self._basename = None;
    self._modified = False;
    self._closed = False;
    self._error_at_line = {};
    self._is_tree_in_sync = True;

  def __del__ (self):
    self.has_focus(False);
    
  def disable_editor (self):
    """Called before disabling the editor, as on some versions of PyQt
    the object is not destroyed properly and keeps receving signals""";
    self._enabled = False;
    
  def show_compile_options (self):
    self._options_menu.show();
    
  def show_runtime_options (self):
    self._jobmenu.show();
    
  def tree_is_in_sync (self,sync=True):
    """Tells the editor wheether the current tree is in sync with the content of the script.
    This is indicated by a visual cue on the toolbar.
    """;
    QToolTip.remove(self._tb_jobs);
    if sync:
      self._tb_jobs.setIconSet(pixmaps.gear.iconset());
      QToolTip.add(self._tb_jobs,"Access run-time options & jobs defined by this TDL script");
    else:
      self._tb_jobs.setIconSet(pixmaps.exclaim_yellow_warning.iconset());
      QToolTip.add(self._tb_jobs,"""Access run-time options & jobs defined by this TDL script.
Warning! You have modified the script since it was last compiled, so the tree may be out of date.""");
  
  def _file_closed (self):
    self.emit(PYSIGNAL("fileClosed()"),(self,));

  def hideEvent (self,ev):
    self.emit(PYSIGNAL("hidden()"),());
    self.emit(PYSIGNAL("visible()"),(False,));
    return QFrame.hideEvent(self,ev);

  def showEvent (self,ev):
    self.emit(PYSIGNAL("shown()"),());
    self.emit(PYSIGNAL("visible()"),(True,));
    return QFrame.showEvent(self,ev);

  def hide_jobs_menu (self,dum=False):
    if self._closed:
      return;
    self._tb_jobs.hide();
    self.clear_message();

  def show_line_numbers (self,show):
    if show:
      self._editor.setMarginWidth(1,36);
    else:
      self._editor.setMarginWidth(1,12);
    self._editor.setMarginLineNumbers(1,show);

  def show_run_control (self,show=True):
    if self._closed:
      return;
    self._tb_run.setShown(show);

  def enable_controls (self,enable=True):
    if self._closed:
      return;
    self._tb_run.setEnabled(enable);
    self._tb_jobs.setEnabled(enable);
    if not enable:
      self.clear_message();

  def disable_controls (self,disable=True):
    if self._closed:
      return;
    self._tb_run.setDisabled(disable);
    self._tb_jobs.setDisabled(disable);
    if disable:
      self.clear_message();

  def get_filename (self):
    return self._filename;
  def get_mainfile (self):
    return self._mainfile;

  def _import_main_file (self):
    # self._tb_opts.setOn(False);
    self.clear_errors();
    if self._mainfile and self._editor.isModified():
      self._save_file();
    self.emit(PYSIGNAL("importFile()"),(self,self._mainfile or self._filename,));

  def _import_as_main_file (self):
    self.clear_errors();
    self._set_mainfile(None);
    self._text_modified(self._editor.isModified());   # to reset labels
    self.emit(PYSIGNAL("fileChanged()"),(self,));
    self.emit(PYSIGNAL("importFile()"),(self,self._filename,));
    
  def _run_main_file (self):
    # self._tb_opts.setOn(False);
    self.clear_errors();
    if self._mainfile and self._editor.isModified():
      self._save_file();
    self.emit(PYSIGNAL("compileFile()"),(self,self._mainfile or self._filename,));

  def _clear_transients (self):
    """if message box contains a transient message, clears it""";
    if self._message_transient:
      self.clear_message();

  def _text_changed (self):
    self._clear_transients();
#    self._qa_run.setVisible(True);

  def _display_cursor_position (self,line,col):
    self._poslabel.setText("L:<b>%d</b> C:<b>%d</b>" % (line+1,col+1));
    self._poslabel.repaint();

  def _text_modified (self,mod):
    self._modified = mod;
    self.emit(PYSIGNAL("textModified()"),(self,bool(mod),));
    if mod:
      self._tb_save.setPaletteBackgroundColor(self._tb_save._modified_color);
    else:
      self._tb_save.unsetPalette();
    if self._filename:
      label = '<b>' + self._basename + '</b>';
      QToolTip.add(self._pathlabel,self._filename);
    else:
      label = '';
      QToolTip.remove(self._pathlabel);
    if self._mainfile:
      label += ' (from <b>' + self._mainfile_base +'</b>)';
    if self._readonly:
      label = '[r/o] ' + label;
    if mod:
      self._clear_transients();
      label = '[mod] ' + label;
    self._pathlabel.setText(label);

  def clear_message (self):
    # print "******* clear_message";
    # traceback.print_stack();
    self._message_box.hide();
    self._message.setText('');
    self._message_icon.setText('');
    self._message_transient = False;
    if self._message_widgets:
      dum = QWidget();
      for w in self._message_widgets:
        w.reparent(dum);
      self._message_widgets = [];

  def show_message (self,msg,error=False,iconset=None,transient=False):
    """Shows message in box.
    If iconset is not None, overrides standard icon.
    If error=True, uses error icon (iconset overrides this).
    If transient=True, message will be cleared when text is edited.
    """;
    self._message.setText(msg);
    if not iconset:
      if error:
        iconset = pixmaps.red_round_cross.iconset();
      else:
        iconset = pixmaps.info_blue_round.iconset();
    self._message_icon.setIconSet(iconset);
    self._message_box.show();
    self._message_transient = transient;
    self.clear_errors();

  def messagebox ():
    return self._message_box;

  def add_message_widget (self,widget):
    self._mblo.addWidget(widget);
    self._message_widgets.append(widget);

  def clear_errors (self):
    self._error_window.clear_errors(emit_signal=True);

  def _reset_errors (self,nerr):
    """helper method, resets error markers and such. Usually tied to a hasErrors() signal
    from an error window""";
    if not self._enabled:
      return;
    self._editor.markerDeleteAll(self.ErrorMarker);
    self._editor.markerDeleteAll(self.SubErrorMarker);
    self._editor.markerDeleteAll(self.CurrentErrorMarker);
    self._error_at_line = {};
    nerr_local = 0;
    if nerr:
      error_locations = self._error_window.get_error_locations();
      for err_num,filename,line,column,suberror in error_locations:
        if filename == self._filename:
          self._error_at_line[line-1] = err_num;
          if suberror:
            self._editor.markerAdd(line-1,self.SubErrorMarker);
          else:
            self._editor.markerAdd(line-1,self.ErrorMarker);
          nerr_local += 1;
    self.emit(PYSIGNAL("hasErrors()"),(self,nerr_local,));

  def show_error (self,err_num,filename,line,column):
    """Shows error at the given position, but only if the filename matches.
    Can be directly connected to a showError() signal from an error window""";
    self._editor.markerDeleteAll(self.CurrentErrorMarker);
    if filename == self._filename:
      # scintilla's line numbers are 0-based
      self._editor.ensureLineVisible(line-1);
      self._editor.setCursorPosition(line-1,column);
      # a little kludge to prevent line from being hidden by a resize
      self._editor.ensureLineVisible(line+4);
      self._editor.markerAdd(line-1,self.CurrentErrorMarker);

  def _process_margin_click (self,margin,line,button):
    _dprint(1,margin,line,button);
    # look through current error widget to find relevant error
    err_num = self._error_at_line.get(line,None);
    if err_num is not None:
      self._error_window.show_error_number(err_num);

  def sync_external_file (self,filename=None,ask=True):
    #
    # NB: problem because it resets the errors
    filename = filename or self._filename;
    filetime = _file_mod_time(filename);
    if not filetime or filetime == self._file_disktime:
      return True;  # in sync
    if not ask:
      res = 1;
    else:
      res = QMessageBox.warning(self,"TDL file changed",
        """<p><tt>%s</tt> has been modified by another program.
        Would you like to overwrite the disk version, revert to the disk
        version, or cancel the operation?"""
        % (filename,),
        "Overwrite","Revert","Cancel",-1,2);
    if res == 2:
      return None;
    elif res == 1:
      if filename != self._filename:
        self.load_file(filename);
      else:
        self.reload_file();
      return True;  # in sync

  def _save_file (self,filename=None,text=None,force=False,save_as=False):
    """Saves text. If force=False, checks modification times in case
    the file has been modified by another program.
    If force=True, saves unconditionally.
    If no filename is known, asks for one.
    Returns True if file was successfully saved, else None.""";
    filename = filename or self._filename;
    if filename and not save_as:
      if not force:
        if not self.sync_external_file(filename=filename,ask=True):
          return None;
    else: # no filename, ask for one
      try: dialog = self._save_as_dialog;
      except AttributeError:
        self._save_as_dialog = dialog = QFileDialog(self,"save tdl dialog",True);
        dialog.resize(800,dialog.height());
        dialog.setMode(QFileDialog.AnyFile);
        dialog.setFilters("TDL scripts (*.tdl);;Python scripts (*.py);;All files (*.*)");
        dialog.setViewMode(QFileDialog.Detail);
        dialog.setCaption("Save TDL Script");
      else:
        dialog.rereadDir();
      if dialog.exec_loop() != QDialog.Accepted:
        return None;
      filename = str(dialog.selectedFile());
    # save the file
    if text is None:
      text = str(self._editor.text());
    try:
      outfile = open(filename,"w").write(text);
    except IOError:
      (exctype,excvalue,tb) = sys.exc_info();
      _dprint(0,'exception',sys.exc_info(),'saving TDL file',filename);
      self.show_message("""<b>Error writing <tt>%s</tt>:
        <i>%s (%s)</i></b>""" % (filename,excvalue,exctype.__name__),
        error=True,transient=True);
      return None;
    # saved successfully, update stuff
    self._filename = filename;
    self._qa_revert.setEnabled(True);
    self._basename = os.path.basename(filename);
    self._file_disktime = _file_mod_time(filename);
    self._editor.setModified(False);
    self._text_modified(False);
    self.emit(PYSIGNAL("fileSaved()"),(self,filename,));
    return self._filename;

  def close (self):
    self._closed = True;

  def confirm_close (self):
    if self._modified:
      res = QMessageBox.warning(self,"TDL file modified",
        """Save modified file <p><tt>%s</tt>?</p>"""
        % (self._filename or "",),
        "Save","Don't Save","Cancel",0,2);
      if res == 2:
        return False;
      if res == 0:
        if not self._save_file():
          return False;
    self.close();
    return True;

  def _revert_to_saved (self,force=False):
    if not self._filename:
      return;
    if not force:
      if QMessageBox.question(self,"Revert to saved",
        """Revert to saved version of <p><tt>%s</tt>?"""
        % (self._filename,),
        "Revert","Cancel","",0,1):
        return;
    self.load_file(self._filename);

  def _add_menu_label (self,menu,label):
    tlab = QLabel("<b>"+label+"</b>",menu);
    tlab.setAlignment(Qt.AlignCenter);
    tlab.setFrameShape(QFrame.ToolBarPanel);
    tlab.setFrameShadow(QFrame.Sunken);
    menu.insertItem(tlab);
    
  def has_compile_options (self):
    return self._options_menu.listView().childCount();
  
  def has_runtime_options (self):
    return self._jobmenu.listView().childCount();

  def import_content (self,force=False,show_options=False):
    """imports TDL module but does not run _define_forest().
    Depending on autosync/modified state, asks to save or revert.
    If module is already imported, does nothing, unless force=True,
    in which case it imports unconditionally.
    If do_compile=True, proceeds to show compile-time options on success,
    or to compile directly if there are no options
    Return value:
      True on successful import
      None if cancelled by user.
    Import errors are posted to the error dialog.
    """;
    _dprint(1,self._filename,"importing");
    self.clear_message();
    self.clear_errors();
    # change the current directory to where the file is
    # os.chdir(os.path.dirname(self._filename));
    # The Python imp module expects text to reside in a disk file, which is
    # a pain in the ass for us if we're dealing with modified text or text
    # entered on-the-fly. So, either save or sync before proceeding
    global _external_sync;
    if self._editor.isModified() or not self._filename:
      if not self._save_file():
        return None;
    else:
      if not self.sync_external_file(ask=False):
        return None;
    # if we already have an imported module and disk file hasn't changed, skip
    #the importing step
    if force or self._tdlmod is None or self._tdlmod_filetime == self._file_disktime:
      # reset data members
      _dprint(2,self._filename,"emitting signal for 0 compile-time options");
      self.emit(PYSIGNAL("hasCompileOptions()"),(self,0,));
      self._options_menu.hide();
      self._options_menu.clear();
      self._tdlmod = None;
      # get text from editor
      tdltext = str(self._editor.text());
      try:
        tdlmod,tdltext = TDL.Compile.import_tdl_module(self._filename,tdltext);
      # catch import errors
      except TDL.CumulativeError as value:
        _dprint(0,"caught cumulative error, length",len(value.args));
        self._error_window.set_errors(value.args,message="TDL import failed");
        return None;
      except Exception as value:
        _dprint(0,"caught other error, traceback follows");
        traceback.print_exc();
        self._error_window.set_errors([value],message="TDL import failed");
        return None;
      # remember module and nodescope
      self._tdlmod = tdlmod;
      self._tdltext = tdltext;
      self._tdlmod_filetime = self._file_disktime;
      # build options menu from compile-time options
      opt_listview = self._options_menu.listView();
      opts = TDLOptions.get_compile_options();
      if opts:
        # add options
        try:
          TDLOptions.populate_option_treewidget(opt_listview,opts);
        except Exception as value:
          _dprint(0,"error setting up TDL options GUI");
          traceback.print_exc();
          self._error_window.set_errors([value],message="Error setting up TDL options GUI");
          return None;
        # self._tb_opts.show();
        _dprint(2,self._filename,"emitting signal for",len(opts),"compile-time options");
        self.emit(PYSIGNAL("hasCompileOptions()"),(self,len(opts),));
    # success, show options or compile
    if show_options and self.has_compile_options():
      self._options_menu.show();
    return True;
      
  def compile_content (self):
    # import content first, and return if failed
    if not self.import_content(force=True):
      return None;
    _dprint(1,self._filename,"compiling forest");
    # clear predefined functions
    self._tb_jobs.hide();
    # make list of publishing nodes
    pub_nodes = [ node.name for node in meqds.nodelist.iternodes()
                  if node.is_publishing() ];
    # try the compilation
    try:
      QApplication.setOverrideCursor(QCursor(Qt.WaitCursor));
      try:
        (_tdlmod,ns,msg) = \
          TDL.Compile.run_forest_definition(
              meqds.mqs(),self._filename,self._tdlmod,self._tdltext,
              parent=self,wait=False);
      finally:
        QApplication.restoreOverrideCursor();
    # catch compilation errors
    except TDL.CumulativeError as value:
      _dprint(0,"caught cumulative error, length",len(value.args));
      self._error_window.set_errors(value.args,message="TDL import failed");
      return None;
    except Exception as value:
      _dprint(0,"caught other error, traceback follows");
      traceback.print_exc();
      self._error_window.set_errors([value]);
      return None;
    # refresh the nodelist
    meqds.request_nodelist(sync=True);
    # restore publishing nodes
    for name in pub_nodes:
      if name in ns.AllNodes():
        meqds.enable_node_publish_by_name(name,sync=True);
    ### NB: presume this all was successful for now
    # insert node scope into TDL module
    setattr(_tdlmod,'_tdl_nodescope',ns);

    # does the script define an explicit job list?
    joblist = getattr(_tdlmod,'_tdl_job_list',[]);
    if not joblist:
      joblist = [];
      # try to build it from implicit function names
      for (name,func) in _tdlmod.__dict__.items():
        if name.startswith("_tdl_job_") and callable(func):
          joblist.append(func);
    # does the script define a testing function?
    testfunc = getattr(_tdlmod,'_test_forest',None);
    if not callable(testfunc):
      testfunc = getattr(_tdlmod,'test_forest',None);
      if callable(testfunc):
        res = QMessageBox.warning(self,"Deprecated method",
          """Your script contains a test_forest() method. This is deprecated
          and will be disabled in the future. Please rename it to
          _test_forest().
          """,
          QMessageBox.Ok);
    if callable(testfunc):
      joblist.append(testfunc);
    from past.builtins import cmp
    from functools import cmp_to_key
    joblist.sort(key=cmp_to_key(lambda a,b:cmp(str(a),str(b))));

    # create list of job actions
    opts = TDLOptions.get_runtime_options();
    self._jobmenu.clear();
    if joblist or opts:
      self._tb_jobs.setOn(False);
      self._tb_jobs.show();
      if opts:
        self._job_executor = curry(self.execute_tdl_job,_tdlmod,ns);
        ## new style:
        try:
          TDLOptions.populate_option_treewidget(self._jobmenu.listView(),opts,executor=self._job_executor);
        except Exception as value:
          _dprint(0,"error setting up TDL options GUI");
          traceback.print_exc();
          self._error_window.set_errors([value],message="Error setting up TDL options GUI");
          return None;
      if joblist:
        for func in joblist:
          name = re.sub("^_tdl_job_","",func.__name__);
          name = name.replace('_',' ');
          self._jobmenu.addAction(name,
              curry(self.execute_tdl_job,_tdlmod,ns,func,name),
              icon=pixmaps.gear);
      self.emit(PYSIGNAL("hasRuntimeOptions()"),(self,True));
    else:
      self.emit(PYSIGNAL("hasRuntimeOptions()"),(self,False));
      self._tb_jobs.hide();

    if joblist:
      msg += " %d predefined function(s) available, please use the Exec menu to run them." % (len(joblist),);

    self.show_message(msg,transient=True);
    return True;

  def execute_tdl_job (self,_tdlmod,ns,func,name):
    """executes a predefined TDL job given by func""";
    self._jobmenu.hide();
    try:
      # log job 
      TDLOptions.dump_log("running TDL job '%s'"%name);
      QApplication.setOverrideCursor(QCursor(Qt.WaitCursor));
      try:
        func(meqds.mqs(),self);
      finally:
        QApplication.restoreOverrideCursor();
      # no errors, so clear error list, if any
      self.clear_errors();
      self.show_message("TDL job '"+name+"' executed.",transient=True);
    except:
      (etype,exc,tb) = sys.exc_info();
      _dprint(0,'exception running TDL job',func.__name__);
      traceback.print_exception(etype,exc,tb);
      # use TDL add_error() to process the error, since this automatically
      # adds location information. However, we want to remove ourselves
      # from the stack traceback first
      tb = traceback.extract_tb(tb);
      # pop everything leading up to our filename
      while tb[0][0] != _MODULE_FILENAME:
        tb.pop(0);
      # pop frame with our filename, this should leave only TDL-code frames
      tb.pop(0);
      ns.AddError(exc,tb,error_limit=None);
      msg = "TDL job '"+name+"' failed";
      self._error_window.set_errors(ns.GetErrors(),signal=True,message=msg);
      self.emit(PYSIGNAL("showEditor()"),(self,));

  def get_jobs_popup (self):
    return self._jobmenu;

  def get_options_popup (self):
    return self._options_menu;

  def _set_mainfile (self,mainfile):
    """adjusts GUI controls based on whether we are a mainfile or not""";
    self._mainfile = mainfile;
    self._mainfile_base = mainfile and os.path.basename(mainfile);
    # if we have a mainfile, add extra options for the Run button
    if mainfile:
      self._tb_run.setPopup(self._tb_runmenu);
      QToolTip.add(self._tb_run,"Saves this script and runs the main script "+self._mainfile_base+". Click on the down-arrow for other options.");
      self._qa_runmain.setToolTip("Saves this script and runs the main script "+self._mainfile_base+".");
      self._qa_runmain.setMenuText("Run "+self._mainfile_base);
    else:
      self._tb_run.setPopup(None);
      QToolTip.add(self._tb_run,"Saves and runs the script.");
      
  def reload_file (self):
    text = open(self._filename).read();
    # set save icons, etc.
    self._qa_revert.setEnabled(True);
    self._file_disktime = _file_mod_time(self._filename);
    self._editor.setText(text);
    self._editor.setReadOnly(not os.access(self._filename,os.W_OK));
    self._editor.setModified(False);
    self._text_modified(False);

  def load_file (self,filename,text=None,readonly=False,mainfile=None):
    """loads editor content.
    filename is filename. text is file contents, if none then file will be re-read.
    readonly is True for readonly mode.
    If mainfile is not None, then this editor is "slaved" to the mainfile. This is the
    case for files included from other modules.
    """
    self.clear_message();
    if not os.access(filename,os.W_OK):
      readonly = True;
    # load text from file if not supplied
    if text is None:
      text = open(filename).read();
    self._filename = filename;
    # sets as as the mainfile or as a submodule of a main file
    self._set_mainfile(mainfile);
    # set save icons, etc.
    self._qa_revert.setEnabled(True);
    self._basename = os.path.basename(filename);
    self._readonly = readonly;
    self._file_disktime = filename and _file_mod_time(filename);
    self._editor.setText(text);
    self._editor.setReadOnly(readonly);
    self._editor.setModified(False);
    self._text_modified(False);
    # emit signals
    self.emit(PYSIGNAL("fileLoaded()"),(self,filename,));
    # if module is a main-level file (i.e. not slaved to another mainfile),
    # pre-import it so that compile-time menus become available
    self._tdlmod = None;
    if not mainfile:
      self.import_content();

  def adjust_editor_font (self):
    self._editor.setFont(self.font());
    self._lexer.setDefaultFont(self.font());
    self._lexer.setFont(self.font(),-1);
    ps = self.fontInfo().pointSize()+self._editor_fontadjust;
    self._editor.zoomTo(ps);

  def has_focus (self,focus):
    if focus:
      TDLEditor.current_editor = self;
    else:
      if TDLEditor.current_editor == self:
        TDLEditor.current_editor = None;


class TDLOptionsDialog (QDialog,PersistentCurrier):
  """implements a floating window for TDL options""";
  def __init__ (self,parent,ok_label=None,ok_icon=None):
    QDialog.__init__(self,parent,"options",False);
    self.setSizeGripEnabled(True);
    
    lo_main = QVBoxLayout(self,11,6);
    # add option listview
    self._listview = listview = QListView(self);
    listview.addColumn("");
    listview.addColumn("");
    listview.addColumn("");
    listview.setRootIsDecorated(True);
    listview.setShowToolTips(True);
    listview.setSorting(-1);
    listview.setResizeMode(QListView.LastColumn);
    # listview.header().hide();
    # listview.setSizePolicy(QSizePolicy.MinimumExpanding,QSizePolicy.MinimumExpanding);
    # add signals
    QObject.connect(listview,SIGNAL("clicked(QListViewItem*)"),self._process_listview_click);
    QObject.connect(listview,SIGNAL("pressed(QListViewItem*)"),self._process_listview_press);
    QObject.connect(listview,SIGNAL("returnPressed(QListViewItem*)"),self._process_listview_click);
    QObject.connect(listview,SIGNAL("spacePressed(QListViewItem*)"),self._process_listview_click);
    QObject.connect(listview,SIGNAL("doubleClicked(QListViewItem*,const QPoint &, int)"),
                            self._process_listview_click);
    # listview.setMinimumSize(QSize(200,100));
    lo_main.addWidget(listview);
    # set geometry
    # self.setMinimumSize(QSize(200,100));
    # self.setGeometry(0,0,200,60);
    
    # add buttons on bottom
    lo_btn  = QHBoxLayout(None,0,6);
    if ok_label:
      if ok_icon:
        tb = QPushButton(ok_icon.iconset(),ok_label,self);
      else:
        tb = QPushButton(ok_label,self);
      lo_btn.addWidget(tb);
      QObject.connect(tb,SIGNAL("clicked()"),self.accept);
    # add cancel button
    tb = QPushButton(pixmaps.red_round_cross.iconset(),"Cancel",self);
    QObject.connect(tb,SIGNAL("clicked()"),self.hide);
    spacer = QSpacerItem(40,20,QSizePolicy.Expanding,QSizePolicy.Minimum);
    lo_btn.addItem(spacer)
    lo_btn.addWidget(tb);
    lo_main.addLayout(lo_btn);
    self.resize(QSize(600,480).expandedTo(self.minimumSizeHint()))
    
  def accept (self):
    self.emit(PYSIGNAL("accepted()"),());
    self.hide();
    
  def show (self):
    self._listview.adjustColumn(0);
    self._listview.adjustColumn(1);
    return QDialog.show(self);
  
  def exec_loop (self):
    self._listview.adjustColumn(0);
    self._listview.adjustColumn(1);
    return QDialog.exec_loop(self);
  
  def listView (self):
    return self._listview;
  
  def clear (self):
    self._listview.clear();
    
  class ActionListViewItem (QListViewItem):
    def paintCell (self,qp,cg,column,width,align):
      # use bold font for jobs
      oldfont = qp.font();
      font = QFont(oldfont);
      font.setBold(True);
      qp.setFont(font);
      QListViewItem.paintCell(self,qp,cg,column,width,align);
      qp.setFont(oldfont);
    def width (self,fm,lv,c):
      return int(QListViewItem.width(self,fm,lv,c)*1.2);
    
  def addAction (self,label,callback,icon=None):
    # find last item
    last = self._listview.firstChild();
    while last and last.nextSibling():
      last = last.nextSibling();
    item = self.ActionListViewItem(self._listview,last);
    item.setText(0,label);
    if icon:
      item.setPixmap(0,icon.pm());
    item._on_click = callback;
    item._close_on_click = True;
    self._listview.adjustColumn(0);
    self._listview.adjustColumn(1);
    
  def _process_listview_click (self,item,*dum):
    """helper function to process a click on a listview item. Meant to be connected
    to the clicked() signal of a QListView""";
    on_click = getattr(item,'_on_click',None);
    if on_click:
      on_click();
    if getattr(item,'_close_on_click',False):
      self.hide();
      
  def _process_listview_press (self,item,*dum):
    """helper function to process a press on a listview item. Meant to be connected
    to the pressed() signal of a QListView""";
    on_click = getattr(item,'_on_press',None);
    if on_click:
      on_click();

class TDLErrorFloat (QMainWindow,PersistentCurrier):
  """implements a floating window for TDL error reports""";
  def __init__ (self,parent):
    fl = Qt.WType_TopLevel|Qt.WStyle_Customize;
    fl |= Qt.WStyle_DialogBorder|Qt.WStyle_Title;
    QMainWindow.__init__(self,parent,"float",fl);
    self.hide();
    self.setIcon(pixmaps.red_round_cross.pm());
    self.setCaption("TDL Errors");
    # make widgets
    self._werrlist_box = QVBox(self);
    self.setCentralWidget(self._werrlist_box);
    # error list header is a toolbar
    errlist_hdr = QToolBar("TDL errors",self,self._werrlist_box);
    errlist_hdr.setSizePolicy(QSizePolicy.Preferred,QSizePolicy.Fixed);
    # prev/next error buttons
    self._qa_prev_err = QAction(pixmaps.red_leftarrow.iconset(),"Show &previous error",Qt.ALT+Qt.Key_P,self);
    self._qa_prev_err.addTo(errlist_hdr);
    QObject.connect(self._qa_prev_err,SIGNAL("activated()"),self._show_prev_error);
    self._qa_next_err = QAction(pixmaps.red_rightarrow.iconset(),"Show &next error",Qt.ALT+Qt.Key_N,self);
    self._qa_next_err.addTo(errlist_hdr);
    QObject.connect(self._qa_next_err,SIGNAL("activated()"),self._show_next_error);
    # label with error count
    self._error_count_label = QLabel(errlist_hdr);
    errlist_hdr.setStretchableWidget(self._error_count_label);
    self._error_count_label.setSizePolicy(QSizePolicy.Preferred,QSizePolicy.Fixed);
    # error list itself
    self._werrlist = QListView(self._werrlist_box);
    self._werrlist.setSizePolicy(QSizePolicy.MinimumExpanding,QSizePolicy.Preferred);
    QObject.connect(self._werrlist,SIGNAL("currentChanged(QListViewItem*)"),self.curry(self._process_item_click,"currentChanged"));
    QObject.connect(self._werrlist,SIGNAL("clicked(QListViewItem*)"),self.curry(self._process_item_click,"clicked"));
    QObject.connect(self._werrlist,SIGNAL("spacePressed(QListViewItem*)"),self.curry(self._process_item_click,"spacePressed"));
    QObject.connect(self._werrlist,SIGNAL("returnPressed(QListViewItem*)"),self.curry(self._process_item_click,"returnPressed"));
    QObject.connect(self._werrlist,SIGNAL("expanded(QListViewItem*)"),self.curry(self._process_item_click,"expanded"));
    self._werrlist.addColumn('');
    self._werrlist.addColumn('');
    self._werrlist.addColumn('');
    self._werrlist.addColumn('');
    self._werrlist.setSorting(-1);
    self._werrlist.setRootIsDecorated(True);
    self._werrlist.setAllColumnsShowFocus(True);
    self._werrlist.header().hide();
    # size the widget
    self.setMinimumSize(QSize(200,60));
    self.setGeometry(0,0,200,60);
    # anchor and position relative to anchor
    self._anchor_widget = None;
    self._anchor_xy0 = None;
    self._anchor_xy  = 0,0;
    self._current_xy = 0,0;
    self._anchor_ref = 0,0;
    self._anchoring = False;
    # explicit geometry set via anchor, None of none set
    self._geom = None;
    # timer used for move operations
    self._move_timer = QTimer(self);
    # internal state
    self._item_shown = None;
    self._in_show_error_item = False;
    self._error_list = [];

  def closeEvent (self,ev):
    """Window closed: hide and set a flag""";
    self.hide();
    ev.ignore();

  def show (self):
    """Only show the window if we have errors in it""";
    if self._error_list:
      QMainWindow.show(self);

  def moveEvent (self,ev):
    QMainWindow.moveEvent(self,ev);
#    print "moveEvent spontaneous:",ev.spontaneous();
    # ignore move events for some time after a move_anchor() -- these are probably caused
    # by the anchor moving us, and not by the user
    if not self._move_timer.isActive():
      self._current_xy = self.geometry().x(),self.geometry().y();
#      print "current xy is",self._current_xy;

  def showEvent (self,ev):
#    print "showEvent spontaneous:",ev.spontaneous();
    if not ev.spontaneous():
      x,y = self._current_xy;
      self.setGeometry(x,y,self.width(),self.height());
      self._move_timer.start(200,True);
    QMainWindow.showEvent(self,ev);
#    print "showEvent ends";

  def set_anchor (self,widget,x,y,xref=0,yref=0):
    """Tells the window to anchor itself to point x,y of the given widget.
    If xref is 0, x is relative to the left side, otherwise to the right side.
    If yref is 0, y is relative to the top edge, otherwise to the bottom edge.
    """;
    self._anchor_widget = widget;
    self._anchor_xy = x,y;
    self._anchor_ref = xref,yref;
#    print "anchoring to",x,y,xref,yref;
    self.move_anchor();

  def move_anchor (self):
    """Notifies the window that its anchor widget has moved around.
    Changes position following the anchor.""";
    # get dx,dy coordinates relative to old anchor point
    if self._anchor_xy0 is not None:
      x0,y0 = self._anchor_xy0;
      x,y = self._current_xy;
#      print "move_anchor: old location is",x,y;
      dx = x - x0;
      dy = y - y0;
    else:
      dx = dy = 0;
#    print "move_anchor: dxy relative to old anchor is",dx,dy;
    # compute new anchoring point
    top = self._anchor_widget.mapToGlobal(QPoint(0,0));
    btm = self._anchor_widget.mapToGlobal(QPoint(self._anchor_widget.width(),
                          self._anchor_widget.height()));
    #print "move_anchor: widget top is",top.x(),top.y();
    #print "move_anchor: widget btm is",btm.x(),btm.y();
    x,y = self._anchor_xy;
    # add to coordinates of reference point on anchor
    if self._anchor_ref[0]:
      x0 = btm.x() + x;
    else:
      x0 = top.x() + x;
    if self._anchor_ref[1]:
      y0 = btm.y() + y;
    else:
      y0 = top.y() + y;
    # move to new location
    #print "move_anchor: new location is",x0,y0;
    self._anchor_xy0 = x0,y0;
    self._current_xy = x0+dx,y0+dy;
    self.setGeometry(x0+dx,y0+dy,self.width(),self.height());
    # Start a timer to ignore move events for a bit. The reason for this is that sometimes
    # very rapid moving of the main window causes the float to "lag" because the setGeometry()
    # call is not processed before another call to move_anchor(). A small delay should eliminate this.
    self._move_timer.start(200,True);

  def _populate_error_list (self,parent,errlist,toplevel=False):
    """helper function to recursively populate the error ListView""";
    previtem = None;
    for index,err in enumerate(errlist):
      if isinstance(err,TDL.TDLError):
        errmsg = str(err.args[0]);
      else:
        errmsg = str(err);
      filename = getattr(err,'filename',None);
      line = getattr(err,'lineno',0);
      column = getattr(err,'offset',0) or 0;  # offset is sometimes None, make sure this is 0
      suberror = isinstance(err,TDL.NestedTDLError);
      _dprint(1,errmsg,"at",filename,line,column);
      # index of corresponding top-level item
      toplevel_index = len(self._toplevel_error_items)-1;
      if toplevel:
        toplevel_index += 1;
      # create item
      if previtem is not None:
        previtem = item = QListViewItem(parent,previtem,"%d:"%(toplevel_index+1,));
      else:
        previtem = item = QListViewItem(parent,"%d:"%(toplevel_index+1,));
      # add housekeeping info
      item._toplevel_index = toplevel_index;
      if toplevel:
        self._toplevel_error_items.append(item);
      toplevel_index = len(self._toplevel_error_items)-1;
      self._error_items.append(item);
      item.setOpen(False);
      # set item content
      if filename is None:
        item.setText(1,errmsg);
        item.setText(2,"<unknown>");
        item.setText(3,"(%s)"%err.__class__.__name__);
      else:
        # normalize filenames: eliminate CWD, leave just one path element
        fname = filename;
        try: is_cwd = os.path.samefile(os.path.dirname(fname),'.');
        except: is_cwd = False;
        if is_cwd:
          fname = os.path.basename(filename);
        else:
          dirname,fname = os.path.split(filename);
          dir1,dir2 = os.path.split(dirname);
          if dir1 == '/' or not dir1:
            fname = filename;
          else:
            fname = os.path.join("...",dir2,fname);
        item.setText(1,errmsg);
        item.setText(3,"(%s)"%err.__class__.__name__);
        item._err_location = len(self._error_items)-1,filename,line,column,suberror;
        self._error_locations.append(item._err_location);
        item.setText(2,"[%s:%d]" % (fname,line));
      # recursively populate with nested errors if needed
      nested = getattr(err,'nested_errors',None);
      if nested:
        self._populate_error_list(item,nested);

  def set_errors (self,error_list,signal=True,show_item=True,
                      message="TDL compile failed"):
    """Shows an error list. errlist should be a sequence of Exception
    objects following the TDL error convention.
    message is a status message
    If show_item=True, highlights the first error
    If signal=True, emits a hasErrors(nerr) pysignal
    """;
    _dprint(1,"set_errors: list of",len(error_list),"entries");
    _dprint(1,"current list has",len(self._error_list),"entries");
    # do nothing if already set
    if self._error_list is error_list:
      return;
    self.clear_errors(emit_signal=False);
    self._error_list = error_list;
    self._error_locations = [];         # list of error locations
    self._error_items = [];
    self._toplevel_error_items = [];
    if error_list:
      self._populate_error_list(self._werrlist,error_list,toplevel=True);
      nerr = len(self._toplevel_error_items);
      self._error_count_label.setText('%s: <b>%d</b> errors'%(message,nerr));
      self.setCaption("TDL Errors: %d"%nerr);
      self.show();
      if signal:
        self.emit(PYSIGNAL("hasErrors()"),(nerr,));
      if show_item:
        self._show_error_item(self._toplevel_error_items[0]);
      # resize ourselves according to number of errors
      height = (len(self._error_items)+1)*self._werrlist.fontMetrics().lineSpacing();
      height = min(200,height);
      self.setGeometry(self.x(),self.y(),self.width(),height);
      self.updateGeometry();
      # self._highlight_error(0);
      # disable run control until something gets modified
      # self._qa_run.setVisible(False);
    else:
      self.setCaption("TDL Errors");
      self.hide();

  def get_error_list (self):
    return self._error_list;

  def get_error_locations (self):
    return self._error_locations;

  def clear_errors (self,signal=True):
    """clears the error list. If signal=True, emits a hasErrors(0) pysignal""";
    if signal:
      self.emit(PYSIGNAL("hasErrors()"),(0,));
    self._error_items = self._toplevel_error_items = None;
    self.setCaption("TDL Errors");
    self._werrlist.clear();
    self._error_count_label.setText('');
    self._error_list = [];
    self._error_locations = [];
    self._item_shown = None;
    self.hide();

  def _highlight_error_item (self,item,signal=True):
    """highlights the given error item. If signal=True, emits a
    showError(index,filename,line,column) or showError(None) pysignal.
    """;
    if item:
      self.show();
      toplevel_index = item._toplevel_index;
      self._qa_prev_err.setEnabled(toplevel_index > 0);
      self._qa_next_err.setEnabled(toplevel_index < len(self._toplevel_error_items)-1);
      # does item contain a location attribute?
      errloc = getattr(item,'_err_location',None);
      if errloc:
        # indicate location
        _dprint(1,"highlighting error in",errloc);
        if signal:
          self.emit(PYSIGNAL("showError()"),errloc);
        return
    # if we fell through to here, then no error has been shown -- emit appropriate signal
    if signal:
      self.emit(PYSIGNAL("showError()"),(None,None,None,None));

  def _show_next_error (self):
    item = self._werrlist.currentItem();
    if item:
      toplevel_index = item._toplevel_index + 1;
      if toplevel_index >= len(self._toplevel_error_items):
        toplevel_index = 0;
      item = self._toplevel_error_items[toplevel_index];
    else:
      item = self._error_items[0];
    if item:
      self._show_error_item(item);

  def _show_prev_error (self):
    item = self._werrlist.currentItem();
    if item:
      toplevel_index = item._toplevel_index - 1;
      if toplevel_index < 0:
        toplevel_index = len(self._toplevel_error_items)-1;
      item = self._toplevel_error_items[toplevel_index];
    else:
      item = self._toplevel_error_items[-1];
    if item:
      self._show_error_item(item);

  def show_error_number (self,index,signal=True):
    self._show_error_item(self._error_items[index],signal=signal);

  def _show_error_item (self,item,signal=True):
    # do nothing if item is None, or already shown, or if already within this function
    if item is None or self._in_show_error_item or item is self._item_shown:
      return;
    self._in_show_error_item = True;
    try:
      self._item_shown = item;
      self._werrlist.setCurrentItem(item);
      self._werrlist.setSelected(item,True);
      for other in self._toplevel_error_items:
        if other is item:
          other.setOpen(True);
        else:
          other.setOpen(False);
      self._werrlist.ensureItemVisible(item);
      self._highlight_error_item(item,signal=signal);
    finally:
      self._in_show_error_item = False;

  def _process_item_click (self,why,item):
    self._show_error_item(item,signal=True);


class TDLFileDataItem (Grid.DataItem):
  """represents a GridDataItem for a TDL script""";
  def __init__ (self,pathname):
    # read the file (exception propagated outwards on error)
    ff = open(pathname);
    text = ff.read();
    ff.close();
    basename = os.path.basename(pathname);
    # create the item
    udi = '/tdlfile/'+pathname;
    name = basename;
    caption = '<b>'+basename+'</b>';
    desc = 'TDL file '+pathname;
    Grid.DataItem.__init__(self,udi,name=name,caption=caption,desc=desc,data=text,viewer=TDLBrowser,refresh=None);
    # add extra pathname attribute for tdl objects
    self.tdl_pathname = pathname;


class TDLBrowser(browsers.GriddedPlugin):
  _icon = pixmaps.text_tdl;
  viewer_name = "TDL Browser";

  def __init__(self,gw,dataitem,cellspec={},default_open=None,**opts):
    browsers.GriddedPlugin.__init__(self,gw,dataitem,cellspec=cellspec);
    self._wedit = TDLEditor(self.wparent());
    self.set_widgets(self.wtop(),dataitem.caption,icon=self.icon());
    if dataitem.data is not None:
      self.set_data(dataitem);
    QObject.connect(self.wtop(),PYSIGNAL("fontChanged()"),self.wtop().adjust_editor_font);

  def wtop (self):
    return self._wedit;
  def editor (self):
    return self._wedit;

  def set_data (self,dataitem,default_open=None,**opts):
    _dprint(3,'set_data ',dataitem.udi);
    pathname = getattr(dataitem,'tdl_pathname',None);
    self._wedit.load_file(pathname,text=dataitem.data);

  def highlight (self,color=True):
    browsers.GriddedPlugin.highlight(self,color);
    self._wedit.has_focus(bool(color));

Grid.Services.registerViewer(str,TDLBrowser,priority=10,check_udi=lambda x:x.endswith('_tdl'));

# # standard method to register actions in the main menu
# #
# def define_mainmenu_actions (menu,parent):
#   _dprint(1,'defining stream control menu actions');
#   global _qa_next,_qa_prev,_qa_run;
#   _qa_next = QAction(pixmaps.blue_round_rightarrow.iconset(),"&Next error",Qt.CTRL+Qt.Key_N,parent);
#   _qa_next.addTo(menu['TDL']);
#   _qa_next.setEnabled(False);
#   _qa_prev = QAction(pixmaps.blue_round_leftarrow.iconset(),"&Previous error",Qt.CTRL+Qt.Key_P,parent);
#   _qa_prev.addTo(menu['TDL']);
#   _qa_prev.setEnabled(False);
#   _qa_run = QAction(pixmaps.blue_round_reload.iconset(),"&Run script",Qt.CTRL+Qt.Key_R,parent);
#   _qa_run.addTo(menu['TDL']);
#   _qa_run.setEnabled(False);
#



#!/usr/bin/env python3


#% $Id$ 

#
# Copyright (C) 2002-2007
# ASTRON (Netherlands Foundation for Research in Astronomy)
# and The MeqTree Foundation
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands, seg@astron.nl
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import sys
import numpy
import math

try:
  from Timba.utils import verbosity
  _dbg = verbosity(0,name='SolverData');
  _dprint = _dbg.dprint;
  _dprintf = _dbg.dprintf;
except:
  pass

class SolverData:
   """ a class to store solver data and supply the
       data to the browser display """

   def __init__(self, data_label=''):
     self._data_label = data_label
     self._solver_array = None
     self.metrics_rank = None
     self.metrics_covar = None
     self.condition_numbers = None 
     self.cn_chi = None 
     self.metrics_chi_0 = None
     self.solver_offsets = None
     self.solver_labels = None
     self.metrics_unknowns = None
     self.chi_array = None
     self.eigenvectors = None
     self.vector_sum = None
     self.incr_soln_norm = None
     self.sum_incr_soln_norm = None
     self.nonlin = None
     self.prev_unknowns = 0
     self.iteration_number = None
#    self.__init__

   def StoreSolverData(self, incoming_data, data_label=''):
     """ This method stores solver data and calculates various
         arrays that describe the behaviour of the solver solutions
         as a function of iteration number
     """
     self.metrics_covar = None
     self._data_label = data_label
     if "incremental_solutions" in incoming_data.solver_result:
       self._solver_array = incoming_data.solver_result.incremental_solutions
       self.chi_array = self._solver_array.copy()
       shape = self._solver_array.shape
       #shape[0] = number of interations == num_metrics (see below)
       #shape[1] = total number of solution elements
       if "metrics" in incoming_data.solver_result:
         metrics = incoming_data.solver_result.metrics
# find out how many records in each metric field
         num_metrics = len(metrics)
         num_metrics_rec =  len(metrics[0])
         self.solver_offsets = numpy.zeros((num_metrics_rec), numpy.int32)
         self.metrics_rank = numpy.zeros((num_metrics,num_metrics_rec), numpy.int32)
         self.metrics_unknowns = numpy.zeros((num_metrics,num_metrics_rec), numpy.int32)
         self.metrics_covar = []
         self.metrics_chi_0 = numpy.zeros((num_metrics,num_metrics_rec), numpy.float64)
         self.metrics_chi = numpy.zeros((num_metrics,num_metrics_rec), numpy.float64)
         self.metrics_fit = numpy.zeros((num_metrics,num_metrics_rec), numpy.float64)
         self.metrics_mu = numpy.zeros((num_metrics,num_metrics_rec), numpy.float64)
         self.metrics_stddev = numpy.zeros((num_metrics,num_metrics_rec), numpy.float64)
         self.metrics_flag = numpy.zeros((num_metrics,num_metrics_rec), numpy.bool_)
         self.vector_sum = numpy.zeros((num_metrics,num_metrics_rec), numpy.float64)
         self.incr_soln_norm = numpy.zeros((num_metrics,num_metrics_rec), numpy.float64)
         self.sum_incr_soln_norm = numpy.zeros((num_metrics,num_metrics_rec), numpy.float64)
         self.iteration_number = numpy.zeros((num_metrics), numpy.int32)
         for i in range(num_metrics):
           if i > 0:
             for j in range(shape[1]):
               self.chi_array[i,j] = self.chi_array[i,j] + self.chi_array[i-1,j]
           self.prev_unknowns = 0
           # we just reset the following list for each iteration, so that
           # only the last one is actually used
           covar_list = []
           for j in range(num_metrics_rec):
             metrics_rec =  metrics[i][j]
             try:
               if "covar" in metrics_rec:
                 covar_list.append(metrics_rec.covar)
               self.metrics_chi_0[i,j] = metrics_rec.chi_0 
               self.metrics_fit[i,j] = metrics_rec.fit 
               self.metrics_chi[i,j] = metrics_rec.chi 
               self.metrics_mu[i,j] = metrics_rec.mu 
               self.metrics_flag[i,j] = metrics_rec.flag 
               self.metrics_stddev[i,j] = metrics_rec.stddev 
               sum_array = 0.0
               sum_array_test = 0.0
               for k in range(self.prev_unknowns,self.prev_unknowns + metrics_rec.num_unknowns):
                  sum_array_test = sum_array_test + self.chi_array[i,k] * self.chi_array[i,k]
                  self.vector_sum[i,j] = self.vector_sum[i,j] + self.chi_array[i,k] * self.chi_array[i,k]
                  sum_array = sum_array + self._solver_array[i,k] * self._solver_array[i,k]
               self.vector_sum[i,j] = math.sqrt(self.vector_sum[i,j])
               self.incr_soln_norm[i,j] = math.sqrt(sum_array)
               if i == 0:
                 self.sum_incr_soln_norm[i,j] = self.sum_incr_soln_norm[i,j] + self.incr_soln_norm[i,j] 
               else:
                 self.sum_incr_soln_norm[i,j] = self.sum_incr_soln_norm[i-1,j] + self.incr_soln_norm[i,j] 
               self.metrics_rank[i,j] = metrics_rec.rank +self.prev_unknowns
               self.prev_unknowns = self.prev_unknowns + metrics_rec.num_unknowns
               self.metrics_unknowns[i,j] = metrics_rec.num_unknowns 
               if i == 0:
                 self.solver_offsets[j] = self.prev_unknowns
             except:
               pass
           self.iteration_number[i] = i+1
         if len(covar_list) > 0:
           self.metrics_covar.append(covar_list)
       if "debug_array" in incoming_data.solver_result:
         debug_array = incoming_data.solver_result.debug_array
# find out how many records in each metric field
         num_debug = len(debug_array)
         num_nonlin =  len(debug_array[0].nonlin)
         self.nonlin = numpy.zeros((num_nonlin, num_debug), numpy.float64)
         for j in range(num_debug):
           debug_rec = debug_array[j]
           nonlin = debug_rec.nonlin
           for i in range(num_nonlin):
             self.nonlin[i,j] = debug_rec.nonlin[i]

# get spid information
       if "spid_map" in incoming_data.solver_result:
         spid_map = incoming_data.solver_result.spid_map
         spid_keys = list(spid_map.keys())
         spid_int_keys = []
         for i in range(len(spid_keys)):
           spid_int_keys.append(int(spid_keys[i]))
         spid_int_keys.sort()
#        print 'spid_map ', spid_map 
#        print 'spid_keys', spid_int_keys
         self.solver_labels = []
         for i in range(len(spid_keys)):
           spid_dict = spid_map[str(spid_int_keys[i])]
           name = ''
           if "name" in spid_dict:
             name = spid_dict["name"]
           coeff_index = ''
           if "coeff_index" in spid_dict:
             coeff_index = spid_dict["coeff_index"]
           label = name + ' ' + str(coeff_index) + ' '
           self.solver_labels.append(label)

   def processCovarArray(self):
     """ get condition number information out of co-variance array """

#    print 'self.metrics_covar ', self.metrics_covar
     self.condition_numbers = []
     self.cn_chi = []
     if self.metrics_covar is None:
       return False
     else:
       if len(self.metrics_covar)== 0:
         return False
       else:
         shape=self.metrics_chi.shape
         num_iter = len(self.metrics_covar)
         # just process the final record
         covar_list = self.metrics_covar[num_iter-1]
         num_covar_matrices = len(covar_list)
         for i in range(num_covar_matrices):
           covar = covar_list[i]
           if covar.min() == 0.0 and covar.max() == 0.0:
             self.condition_numbers.append(None)
             self.cn_chi.append(None)
           else:
             # following equation provided by Sarod
             cond_number=numpy.linalg.norm(covar,2)/numpy.linalg.norm(covar,-2);
             self.condition_numbers.append(cond_number)
             self.cn_chi.append(cond_number * self.metrics_chi[shape[0]-1,i])
         return True

   def calculateCovarEigenVectors(self):
     """ calculate eigenvalues and eigenvectors of co-variance matrix """

     try:
       import numpy.linalg as la
     except:
       return False
     if self.metrics_covar is None:
       return False
     else:
       if len(self.metrics_covar)== 0:
         return False
       else:
         num_iter = len(self.metrics_covar)
         # just process the final record
         covar_list = self.metrics_covar[num_iter-1]
         num_covar_matrices = len(covar_list)
         self.eigenvectors = []
         for i in range(num_covar_matrices):
           covar = covar_list[i]
           shape = covar.shape
           if covar.min() == 0.0 and covar.max() == 0.0:
             self.eigenvectors.append(None)
           elif shape[0] != shape[1]:
             self.eigenvectors.append(None)
           else:
             # gets both eigenvalues and eigenvectors
             self.eigenvectors.append(la.eig(covar))
             # for moment, just get eigenvalues
#            self.eigenvectors.append(la.eigvals(covar))
         return True

   def getConditionNumbers(self):
     """ return the covariance matrix condition numbers """
     return (self.condition_numbers,self.cn_chi)

   def getEigenVectors(self):
     """ return the eigenvalues and eigenvectors of the covariance matrix """
     return self.eigenvectors

   def getSolverLabels(self):
     """ return the solver labels for the display """
     return self.solver_labels

   def getSolverData(self):
     """ return the solver incremental solutions array """
     return self._solver_array

   def getSolverMetrics(self):
     """ return a tuple that contains the various solver metrics
         arrays that are plotted in the chi_sq surface display
     """
     return (self.metrics_rank, self.iteration_number, self.solver_offsets, self.vector_sum, self.metrics_chi_0, self.nonlin, self.sum_incr_soln_norm, self.incr_soln_norm, self.metrics_fit, self.metrics_chi, self.metrics_mu, self.metrics_flag, self.metrics_stddev,self.metrics_unknowns)

def main(args):
  print('we are in main') 

# Admire
if __name__ == '__main__':
    main(sys.argv)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#
#% $Id$ 
#
#
# Copyright (C) 2002-2007
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from Timba.utils import *
from MeqGUI.GUI.pixmaps import pixmaps
from MeqGUI.GUI.widgets import *
from Timba import *
from MeqGUI import GUI, Plugins, Grid
from MeqGUI.Grid.Debug import *
import MeqGUI.Grid.Services

import weakref
import re
import gc
import types

from PyQt4.Qt import *
import Kittens
from Kittens.widgets import PYSIGNAL

# ====== Grid.Cell ==============================================================
# manages one cell of a gridded workspace
#
class Cell (object):
  """The Cell class represents a grid cell with float/pin/close/menu
  controls and a title label.
  
  The top-level widget of a cell is available as cell.wtop(). This
  widget will emit the following PYSIGNALS:
    "clicked()"             : cell has been clicked anywhere
    "float()"               : float of cell is requested
    "flash(bool)"           : cell is flash/unflashing its refresh button  
    "highlight(color|bool)" : cell is highlighted (True for default color)
                              or dehighlighted (False)
    "enable(bool)"          : cell is enabled/disabled
    "wiped()"               : cell contents are cleared
    "closed()"              : cell has been closed via the close button 
                              (wiped() will also be emitted)
    "fontChanged(QFont)"    : cell font has been changed. The same signal is
                              emitted by the content widget.
    "changeViewer(viewer_class)": different viewer selected from "Dispay with" menu
    "itemDropped(item)"     : a dataitem has been dropped on the cell.
                                        
  In addition, a cell will connect PYSIGNAL("refresh()") from its content
  widget to the refresh function of its dataitem, if available. Thus cell
  contents may ask to refresh themselves by emitting this signal. Also,
  when the cell's refresh button is clicked, "refresh()" is emitted on behalf
  of the content widget, thus causing a refresh.
  """
  # define top-level cell widget class. This accepts data drops, 
  # displays context menus, has toolbars, etc.
  DataDroppableQWidget = DataDroppableWidget(QWidget);
  class TopLevelWidget (DataDroppableQWidget):
    def __init__ (self,cell,parent):
      Cell.DataDroppableQWidget.__init__(self,parent);
      self._cell = cell;
      self._menu = None;
    def set_context_menu (self,menu):
      self._menu = menu;
    def mousePressEvent (self,ev):
      self.emit(SIGNAL("clicked"));
      ev.ignore();
    def contextMenuEvent (self,ev):
      # self.emit(SIGNAL("clicked"));
      ev.accept();
      if self._menu:
        self._menu.exec_(ev.globalPos());
    def accept_drop_item_type (self,itemtype):
      return issubclass(itemtype,MeqGUI.Grid.DataItem);
    def get_cell (self):
      return self._cell;
      
  class DraggableCellLabel (QLabel):
    def __init__ (self,cell,*args):
      QLabel.__init__(self,*args);
      self._grid_cell = cell;
      self._pressed = None;
      self._dragtimer = QTimer(self);
      self._dragtimer.setSingleShot(True);
      QObject.connect(self._dragtimer,SIGNAL("timeout()"),self._timeout);
    # these are standard methods to support drags
    def get_drag_item (self,key):
      return self._grid_cell.content_dataitem();
    def get_drag_item_type (self,key):
      item = self._grid_cell.content_dataitem();
      if item is None:
        return None;
      return type(item);
    # process mouse and timer events to start a drag  
    def mousePressEvent (self,ev):
      if ev.button() == Qt.LeftButton \
         and self._grid_cell.content_dataitem() is not None:
        self._pressed = ev.pos();
        self._dragtimer.start(QApplication.startDragTime());
      return QLabel.mousePressEvent(self,ev);
    # mouse released, cancel drag  
    def mouseReleaseEvent (self,ev):
      self._pressed = None;
      self._dragtimer.stop();
      return QLabel.mouseReleaseEvent(self,ev);
    # mouse move, start drag if far enough      
    def mouseMoveEvent (self,ev):
      # start a drag if mouse has moved far enough
      if self._pressed and ev.buttons() & Qt.LeftButton: 
        dist = (self._pressed - ev.pos()).manhattanLength();
        if dist >= QApplication.startDragDistance():
          self._dragtimer.stop();
          self._start_drag();
      return QLabel.mouseMoveEvent(self,ev);
    # timer fired, start drag if mouse still pressed
    def _timeout (self):
      if self._pressed:
        self._start_drag();
    # this method starts the drag  
    def _start_drag (self):
      item = self._grid_cell.content_dataitem();
      if item is not None:
        drag = QDrag(self);
        mimedata = QMimeData();
        mimedata.setUrls([QUrl(item.udi)]);
        drag.setMimeData(mimedata);
        drag.exec_();
        
  def __init__ (self,parent,gridpos,page=None,fixed_cell=False,notitle=False,noviewer=False):
    """constructor. 
    parent:     parent widget
    gridpos:    the grid position. A tuple of (gridpage,row,col).
                This is used to uniquely identify the cell within the workspace.
    noviewer:   if True, the cell will not support change of viewers or drops
                of data items
    fixed_cell: if True, the cell will not support change of viewers, closing,
                pinning, etc., and will not accept drops of data items.
    notitle:    if True, cell will be fixed, and will not have a 'titlebar'.
    page:       this cell's Page parent. None if cell does not belong to a grid
                page (e.g. a floater window)
    """
    # init state
    self._content_widget = self._leader = self._udi = None;
    self._gridpos = gridpos;
    self._refresh_func = lambda:None;
    self._highlight = False;
    self._enabled = False;  
    self._dataitem = None;
    self._parent_page = page;
    # init widgets
    wtop = self._wtop = self.TopLevelWidget(self,parent);
    wtop.hide();
    self._wtop_lo = QVBoxLayout(wtop);
    self._wtop_lo.setSpacing(0);
    self._wtop_lo.setContentsMargins(0,0,0,0);
    self.enable_viewers(not noviewer);
    # --- build toolbar
    self._toolbar = QToolBar("Panel tools",wtop);
    self._toolbar.setIconSize(QSize(16,16));
    self._wtop_lo.addWidget(self._toolbar);
    # icon button and popup menu    
    self._icon_act = self._toolbar.addAction("Panel &menu",self.show_popup_menu);
    self._icon_act.setToolTip("Displays panel menu");
    # refresh button
    self._refresh = refresh = self._toolbar.addAction(pixmaps.refresh.icon(),
				"&Refresh contents",self._dorefresh);
    refresh.setToolTip("refresh contents of this panel");
    self._flashcolor = QColor("yellow");
    # stretchable label
    self._label = self.DraggableCellLabel(self,"(empty)",self._toolbar);
    self._label.setAlignment(Qt.AlignLeft|Qt.AlignVCenter);
    self._label.setSizePolicy(QSizePolicy.Expanding,QSizePolicy.Minimum);
    self._toolbar.addSeparator();
#    qaw = QWidgetAction(wtop);
#    qaw.setDefaultWidget(self._label);
#    self._toolbar.addAction(qaw);
    self._toolbar.addWidget(self._label);
    self._toolbar.addSeparator();
    # pin button
    pin_icon = pixmaps.pin_up.icon();
    pin_icon.addPixmap(pixmaps.pin_down.pm(),QIcon.Normal,QIcon.On);
    self._pin = pin = self._toolbar.addAction(pin_icon,"&Pin/unpin panel");
    pin.setToolTip("pin (i.e. protect) or unpin this panel");
    pin.setCheckable(True);
    # float button
    self._float_act = flt = self._toolbar.addAction(pixmaps.float_window.icon(),"&Float panel",self._float);
    flt.setToolTip("float this cell in a separate window");
    # flt.setToggleAction(True);
    # close button
    self._toolbar.addSeparator();
    self._close = close = self._toolbar.addAction(pixmaps.cancel.icon(),"&Close panel",self.close);
    close.setToolTip("close this panel");
    # finalize toolbar setup
    #self._toolbar.setHorizontallyStretchable(True);
    #self._toolbar.setStretchableWidget(self._label);
    self._toolbar.setMovable(False);
    try: # early Qt4 versions do not have this function
      self._toolbar.setFloatable(False);
    except AttributeError:
      pass;
    self._label.setEnabled(False);
    self._refresh.setVisible(False);
    self._pin.setVisible(False);
    self._close.setVisible(False);
    self._float_act.setVisible(False);
    # set resize policy
    wtop.setSizePolicy(QSizePolicy(QSizePolicy.MinimumExpanding,QSizePolicy.MinimumExpanding));
    # add currier for self.connect() below
    self._currier = PersistentCurrier();
    # clear menus
    self._menu = self._font_menu = self._viewers_menu = None;
    
  def wtop (self):
    return self._wtop;
  def wcontent (self):
    try: return self._content_widget();
    except AttributeError: return None;
  def connect (self,signal,receiver,*args,**kws):
    """connects a signal from the cell's top-level widget to the receiver.
    This is really a kludge to get around the fact that PyQt does not support
    QObject.disconnect() properly, so there's no easy way to disconnect
    all of an object's signals -- which we need to do when clearing a cell.
    Hence, we provide our own connect/disconnect.
    The syntax is similar to the standard utils.curry() call, except with 
    an extra signal argument.
    """;
    # wrap the signal receiver in a curry -- since connect() holds weakrefs,
    # all we need to do for a full disconnect is clear the currier.
    cc = self._currier.curry(receiver,*args,**kws);
    QObject.connect(self._wtop,signal,cc);
    
  def _float (self):
    self.wtop().emit(PYSIGNAL("float()"));
    
  def disconnect_all (self):
    self._currier.clear();
    
  def parent_page (self):
    return self._parent_page;
  def grid_position (self):
    return self._gridpos;
  def hide (self):
    self._wtop.hide();
  def show (self):
    self._wtop.show();
  def enable_viewers (self,enable):
    self._enable_viewers = enable;
    self._wtop.setAcceptDrops(enable);
  def get_udi (self):
    return self._udi;
  def is_parent_of (self,udi):
    return udi and self._udi and len(udi) > len(self._udi) and udi.startswith(self._udi);
  def is_empty (self):
    return self._content_widget is None;
  def is_pinned (self):
    return self._pin.isChecked();
  def set_pinned (self,state=True):
    self._pin.setChecked(state);
  def show_popup_menu (self,point=None):
    self._menu.exec_(point or QCursor.pos());
  def _toggle_pinned_state (self):
    self.set_pinned(not self.is_pinned());
    
  # highlights a cell
  # pass in a QColor, or True for default color, or False value to disable highlights
  def highlight (self,enable=True):
#      wlist = [self._control_box] + self._control_box.children();
#    wlist = (self._toolbar,self._label,self._label1);
    wlist = (self._label,);
    if enable:
      for w in wlist:
        if isinstance(w,QWidget):
          if not self._highlight:
            w._default_background = w.backgroundRole();
            w._default_foreground = w.foregroundRole();
#          w.setBackgroundRole(QPalette.Dark);
          w.setForegroundRole(QPalette.ButtonText);
      self._highlight = True;
    else:
      for w in wlist:
        if isinstance(w,QWidget):
          if hasattr(w,'_default_background'):
            w.setBackgroundRole(w._default_background);
          if hasattr(w,'_default_foreground'):
            w.setForegroundRole(w._default_foreground);
      self._highlight = False;
    # if we're a leader, grab the focus
    if not self._leader:
      self.wtop().setFocus();
    self.wtop().emit(SIGNAL("highlight"),enable,);
      
  def _dorefresh (self):
    if self._content_widget:
      self._content_widget.emit(SIGNAL("refresh"));
    
  def _clear_content (self):
    if self._content_widget:
      self._content_widget.setParent(QWidget());
      self._content_widget = None;
      
  # wipe: deletes contents of cell, possibly in preparation for inserting 
  # other content
  def wipe (self,delete_content=True,close=False):
    """if delete_content==True, content widget is reparented and destroyed,
    and a wiped() signal is emitted. Otherwise, reference to content widget
    is removed but the widget is left alone (this is used for reparenting the
    widget into, e.g., a float window).
    If close==True, a closed() signal is also emitted.
    """;
    _dprint(5,id(self),': wiping cell');
    self._dataitem = self._udi = None;
    self.set_pinned(False);
    self._refresh_func = lambda:None;
    self.wtop().set_context_menu(None);
    self._label.setToolTip('');
    self.highlight(False);
    self._label.setText("(empty)");
    self._label.setEnabled(False);
    self._icon_act.setVisible(False);
    self._refresh.setVisible(False);
    self._pin.setVisible(False);
    self._close.setVisible(False);
    self._float_act.setVisible(False);
    if delete_content:
      self._clear_content();
      _dprint(5,id(self),': emitting wiped() signal');
      self.wtop().emit(SIGNAL("wiped"));
    else:
      self._content_widget = None;
    if close:
      _dprint(5,id(self),': emitting closed() signal');
      self.wtop().emit(SIGNAL("closed"));
    # this is the last step
    self.disconnect_all();

  # close(): wipe, hide everything, and emit a closed signal
  def close (self):
    _dprint(5,id(self),': clearing cell');
    self.wipe(close=True);
    
  def release (self):
    """same as wipe, except content widget is not deleted, and wiped()
    signal is not sent. Used when content has been reparented."""
    self.wipe(delete_content=False);

  def disable (self,disable=True):
    self.enable(not disable);
    
  def enable (self,enable=True):
    self._enabled = enable;
    for w in (self._label,self._close,self._toolbar):
      w.setEnabled(enable);
    if enable:
      self.wtop().set_context_menu(self._menu);
    else:
      self.wtop().set_context_menu(None);
    self.wtop().emit(SIGNAL("enable"),enable,);
    
  def is_enabled (self):
    return self._enabled;
    
  def cell_menu (self):
    return self._menu;
    
  def clear_menu (self):
    self._menu = self._font_menu = self._viewers_menu = None;
    
  # rebuild or clears menu
  def rebuild_menu (self,title,font=None,use_refresh=False):
    # create menu
    self._menu = menu = QMenu(self.wtop());
    if title:
      self._menu_label = Kittens.widgets.addMenuLabel(menu,title);
    if font:
      fontmenu = menu.addMenu(pixmaps.fontsizedown.icon(),"Change font size");
      self._font_qas = {};
      ag = QActionGroup(menu);
      ag.setExclusive(True);
      # populate font menu
      ps = font.pointSize();
      if ps<0:
        ps = font.pixelSize();
      rng = list(range(min(4,ps),15)) + list(range(16,max(26,ps),2));
      for sz in rng:
        self._font_qas[sz] = qa = ag.addAction(str(sz));
        qa.setCheckable(True);
        fontmenu.addAction(qa);
        QObject.connect(qa,SIGNAL("toggled(bool)"),self._menu_currier.curry(self.set_font_size,sz));
      self._reset_font_menu(ps);
    if self._enable_viewers and self._viewers_menu:
      menu.addMenu(self._viewers_menu);
    menu.addAction(self._pin);
    menu.addAction(self._float_act);
    if use_refresh:
      menu.addAction(self._refresh);
    menu.addAction(self._close);
    
  def adjust_font (self,incr):
    """adjusts content font size by 'incr' points""";
    font = self._content_widget.font();
    ps = font.pointSize();
    if ps >= 0:
      self.set_font_size(ps+incr);
    else:
      self.set_font_size(font.pixelSize()+incr);
    
  def set_font_size (self,size,really_set=True):
    """sets content font size. Really_set should be True to actually set the font, this 
    is to make the method compatible with the toggled(bool) signal.""";
    if not really_set:
      return;
    font = QFont(self._content_widget.font());
    if font.pointSize() > 0:
      font.setPointSize(size);
    else:
      font.setPixelSize(size);
    self._content_widget.setFont(font);
    self._content_widget.emit(SIGNAL("fontChanged"),font,);
    self.wtop().emit(SIGNAL("fontChanged"),font,);
    self._reset_font_menu(size);
    
  def _reset_font_menu (self,size):
    qa = self._font_qas.get(size);
    if qa:
      qa.setChecked(True);
      
  def _change_viewer (self,dataitem,viewer):
    self.wtop().emit(SIGNAL("changeViewer"),dataitem,viewer);
    
  def exclusive_highlight (self):
    if self._dataitem:
      MeqGUI.Grid.Services.highlightDataItem(self._dataitem);
    
  def set_content (self,widget,dataitem=None,leader=None,icon=None,enable_viewers=True):
    """inits cell as a leader cell for a dataitem. Leader cells have
    an active toolbar and menu. When a viewer is using a single cell,
    that cell is always a leader cell."""
    # check that widget is our child
    if widget.parent() is not self.wtop():
      raise ValueError('content widget must be child of this Grid.Cell');
    _dprint(3,id(self),': setting content');
    # connect a click on the titlebar to highlight ourselves 
    self.connect(PYSIGNAL("clicked()"),self.exclusive_highlight);
    # init cell content
    self.enable_viewers(enable_viewers);
    self._wtop.hide();
    # insert new content
    self._clear_content();
    self._content_widget = widget;
    widget.setParent(self.wtop());
    self._wtop_lo.addWidget(widget);
    _dprint(5,id(self),': widget',widget,'added');
    # recreate currier for callbacks (discards old curries)
    self._menu_currier = PersistentCurrier();
    self._dataitem = dataitem;
    # setup cell controls based on data item
    if dataitem is not None:
      self._udi = dataitem.udi;
      if leader is not None:
        raise ValueError("both dataitem and leader specified");
      self._leader = None;
      _dprint(5,id(self),': dataitem is',dataitem);
      self.clear_menu();
      # set icon for the menu button
      self._icon_act.setIcon(icon or pixmaps.viewmag.icon());
      # build a "Display with" submenu 
      if len(dataitem.viewer_list) > 1:
        vmenu = self._viewers_menu = QMenu(self.wtop());
        vmenu.setTitle("Display using");
        vmenu.setIcon(pixmaps.viewmag.icon());
        for v in dataitem.viewer_list:
          # create entry for viewer
          name = getattr(v,'viewer_name',v.__name__);
          try: icon = v.icon();
          except AttributeError: icon = QIcon();
          func = self._menu_currier.xcurry(self._change_viewer,_args=(dataitem,v));
          qa = vmenu.addAction(icon,name,func);
          qa.setCheckable(True);
          qa.setChecked(v is dataitem.viewer_obj.__class__);
      else:
        self._viewers_menu = None;
      # build main/context menu
      self.rebuild_menu(dataitem.caption,font=widget.font(),
                        use_refresh=dataitem.is_mutable());
      # setup refresh control
      if dataitem.is_mutable():
        self._refresh_func = dataitem.refresh_func;    
        QWidget.connect(self._content_widget,PYSIGNAL("refresh()"),self._refresh_func);
        self._refresh.setVisible(True);
      else:
        self._refresh_func = lambda:None;
        self._refresh.setVisible(False);
      # setup pin and close button
      self._pin.setVisible(True);
      self._close.setVisible(True);
      self._float_act.setVisible(True);
      # enable cell if data is available
      self.enable(dataitem.data is not None);
    else: # no dataitem, assume follower cell
      if leader is None:
        raise ValueError("leader must be specified if dataitem isn't");
      _dprint(5,id(self),': leader is',leader);
      self._leader = leader;
      self._dataitem = leader._dataitem;
      self._udi = leader.get_udi();
      # disable controls
      self.clear_menu();
      self._refresh.setVisible(False);
      self._pin.setVisible(False);
      self._close.setVisible(False);
      self._float_act.setVisible(False);
      # enable cell as per leader
      self.enable(leader.is_enabled());
      # attach leader's signals
      leader.connect(PYSIGNAL("enable()"),self.enable);
      leader.connect(PYSIGNAL("highlight()"),self.highlight);
      leader.connect(PYSIGNAL("closed()"),self.close);
    self._icon_act.setVisible(True);
    self._label.setEnabled(True);
    # disable cell if no data yet
    # show and resize everything
    self._toolbar.show();
    self._content_widget.show();
    self._wtop.show();
    self._refit_size();
    
  def set_caption (self,caption):
    """sets the caption of a cell."""
    self._label.setToolTip(caption);
    self._label.setText(caption);
    # set menu title if menu is present
    try: self._menu_label.setText(caption);
    except AttributeError: pass;
    
  def leader (self):
    return self._leader;
  def content_udi (self):
    return self._udi;
  def content_dataitem (self):
    return self._dataitem;
    
  def flash (self,flash=True):
    _dprint(5,id(self),': flash()');
    self.wtop().emit(SIGNAL("flash"),flash,);
    self.enable();
    if flash:
      # self._refresh_btn.setPaletteBackgroundColor(self._flashcolor);
      self._refresh.setIcon(pixmaps.refresh_highlight.icon());
      QTimer.singleShot(300,self._unflash);
    else:
      self._unflash();
      
  def _unflash (self):
    # self._refresh_btn.unsetPalette();
    self._refresh.setIcon(pixmaps.refresh.icon());
 
  def _refit_size (self):
    """ugly kludge to get the layout system to do its stuff properly after
    viewer widget has been inserted.""";
    #topwidget = self.wtop().topLevelWidget();
    #sz = topwidget.size();
    #topwidget.resize(sz.width(),sz.height()+1);
    #topwidget.resize(sz);
    pass;



#!/usr/bin/env python3


#% $Id: plot_printer.py 6837 2009-03-05 19:00:25Z twillis $ 

#
# Copyright (C) 2002-2007
# ASTRON (Netherlands Foundation for Research in Astronomy)
# and The MeqTree Foundation
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands, seg@astron.nl
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#  (c) 2013.				 (c) 2011.
#  National Research Council		 Conseil national de recherches
#  Ottawa, Canada, K1A 0R6 		 Ottawa, Canada, K1A 0R6
#
#  This software is free software;	 Ce logiciel est libre, vous
#  you can redistribute it and/or	 pouvez le redistribuer et/ou le
#  modify it under the terms of	         modifier selon les termes de la
#  the GNU General Public License	 Licence Publique Generale GNU
#  as published by the Free		 publiee par la Free Software
#  Software Foundation; either	 	 Foundation (version 3 ou bien
#  version 2 of the License, or	 	 toute autre version ulterieure
#  (at your option) any later	 	 choisie par vous).
#  version.
#
#  This software is distributed in	 Ce logiciel est distribue car
#  the hope that it will be		 potentiellement utile, mais
#  useful, but WITHOUT ANY		 SANS AUCUNE GARANTIE, ni
#  WARRANTY; without even the	 	 explicite ni implicite, y
#  implied warranty of			 compris les garanties de
#  MERCHANTABILITY or FITNESS FOR	 commercialisation ou
#  A PARTICULAR PURPOSE.  See the	 d'adaptation dans un but
#  GNU General Public License for	 specifique. Reportez-vous a la
#  more details.			 Licence Publique Generale GNU
#  					 pour plus de details.
#
#  You should have received a copy	 Vous devez avoir recu une copie
#  of the GNU General Public		 de la Licence Publique Generale
#  License along with this		 GNU en meme temps que ce
#  software; if not, contact the	 logiciel ; si ce n'est pas le
#  Free Software Foundation, Inc.	 cas, communiquez avec la Free
#  at http://www.fsf.org.		 Software Foundation, Inc. au
#						 http://www.fsf.org.
#
#  email:				 courriel:
#  business@hia-iha.nrc-cnrc.gc.ca	 business@hia-iha.nrc-cnrc.gc.ca
#
#  National Research Council		 Conseil national de recherches
#      Canada				    Canada
#  Herzberg Institute of Astrophysics	 Institut Herzberg d'astrophysique
#  5071 West Saanich Rd.		 5071 West Saanich Rd.
#  Victoria BC V9E 2E7			 Victoria BC V9E 2E7
#  CANADA					 CANADA
#
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import printfilter_qt5

class plot_printer:
  def __init__(self, plotter, colorbar=None):

    self.plotter = plotter
    self.colorbar = colorbar
  
# The following 3 functions are adapted from the 'scicraft'
# visualization package

  def _get_qpainter(self, qprinter, hor_widgets, vert_widgets):
    """Returns a qpainter using the given qprinter and geometry."""
    qpainter = Qt.QPainter(qprinter)
    paint_device = qpainter.device()
    width = paint_device.width()
    height = (width // hor_widgets) * vert_widgets
#   qpainter.setClipRect(0, 0, width, height, qpainter.CoordPainter)
    qpainter.setClipRect(0, 0, width, height)
    return qpainter

  def _print_plots(self, qprinter, filter, hor_widgets, vert_widgets, is_complex):
    """Prints all plots with the given qprinter.
    """
    qpainter = self._get_qpainter(qprinter, hor_widgets, vert_widgets)
    # get width and height for each plot 
    paint_device = qpainter.device()
    if hor_widgets > 1:
      if paint_device.width() > paint_device.height():
        # width of plots in x-direction is the largest (wrt. paintdevice)
        width = paint_device.width() // hor_widgets
        height = width # quadratically sized plots
      else:
        # height of plots in x-direction is the largest (wrt. paintdevice)
        height = paint_device.height() // hor_widgets
        width = height # quadratically sized plots
      if not self.colorbar is None:
        try:
          self.colorbar[0].print_(qpainter,
            Qt.QRect(0.02*width, 0, 0.12 * width, height), filter)
#         print 'colorbar plot succeeded'
        except:
          self.colorbar[0].printPlot(qpainter,
            Qt.QRect(0, 0, 0.12 * width, height), filter)
          print('colorbar plot excepted')
        if is_complex:
          try:
            self.colorbar[1].print_(qpainter,
              Qt.QRect(1.6* width, 0, 0.12 * width, height), filter)
          except:
            self.colorbar[1].printPlot(qpainter,
              Qt.QRect(1.6* width, 0, 0.12 * width, height), filter)
      try:
        self.plotter.print_(qpainter,
          Qt.QRect(0.16 * width, 0, 1.4 * width, height), filter)
#       print 'main plot succeeded'
      except:
        self.plotter.printPlot(qpainter,
          Qt.QRect(0.16 * width, 0, 1.4 * width, height), filter)
        print('main plot excepted')
    else:
      width = paint_device.width()
      if width > paint_device.height():
        width =  paint_device.height()
      height = width
      try:
        self.plotter.print_(qpainter,
          Qt.QRect(0, 0, width, height), filter)
      except:
        self.plotter.printPlot(qpainter,
          Qt.QRect(0, 0, width, height), filter)
    qpainter.end()

  def do_print(self, is_single=True, is_complex=False):
    """Sends plots in this window to the printer.
    """
    printer = Qt.QPrinter(Qt.QPrinter.HighResolution)

    printer.setOutputFileName('image-plot-%s.ps' % Qt.qVersion())

#   printer.setCreator('Bode example')
    printer.setOrientation(Qt.QPrinter.Landscape)
    printer.setColorMode(Qt.QPrinter.Color)

    docName = self.plotter.title().text()
    if not docName.isEmpty():
#       docName.replace(Qt.QRegExp(Qt.QString.fromLatin1('\n')), self.tr(' -- '))
        printer.setDocName(docName)

    dialog = Qt.QPrintDialog(printer)
    if dialog.exec_():
        filter = printfilter_qt4.PrintFilter()
        if (Qt.QPrinter.GrayScale == printer.colorMode()):
            filter.setOptions(
                Qwt.QwtPlotPrintFilter.PrintAll
                & ~Qwt.QwtPlotPrintFilter.PrintBackground
                | Qwt.QwtPlotPrintFilter.PrintFrameWithScales)
            filter.setOptions(Qwt.QwtPlotPrintFilter.PrintAll)
# we have 'two' horizontal widgets - colorbar(s) and the display area
        hor_widgets = 2
        if is_single:
          hor_widgets = 1
        self._print_plots(printer, filter, hor_widgets, 1, is_complex)

  def add_colorbar(self, colorbar):
    self.colorbar = colorbar

# class plot_printer



#
#% $Id$ 
#
#
# Copyright (C) 2002-2007
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import Timba
from MeqGUI import Grid, GUI, Plugins
from Timba.dmi import *

from MeqGUI.GUI.pixmaps import pixmaps

from MeqGUI.GUI.treebrowser import NodeAction


import MeqGUI.GUI.app_proxy_gui


from MeqGUI import Grid

from MeqGUI.GUI import browsers

from Timba.ParmDB import *

from qt import *
from qttable import *
import os


_dbg = verbosity(0,name='TableInspector');
_dprint = _dbg.dprint;
_dprintf = _dbg.dprintf;


class TableInspector(browsers.GriddedPlugin):
    viewer_name = "TableInspector";
    _icon = pixmaps.table;
    """GUI to inspect contents of a ParmDB"""
    def is_viewable (data):
        return True;
    
    is_viewable = staticmethod(is_viewable);

    def __init__(self,gw,dataitem,cellspec={},**opts):
        
        browsers.GriddedPlugin.__init__(self,gw,dataitem,cellspec=cellspec,**opts);
        self._db=None;
        self._mainbox = QHBox(self.wparent());
#        self._mainbox = self.wparent();
        self._mainbox.setSpacing(6);
        self._wtop= self._mainbox;

        self._mainvbox = QVBox(self._mainbox);

        #table name menu
        self._tabbox= QVBox(self._mainvbox);
        self._tabbox.setSpacing(6);
        self._table_name=None;
        self._name_label = QLabel("Table Name",self._tabbox);
        self._name_edit = QLineEdit(self._tabbox,"test");
        self._browse_button = QPushButton(self._tabbox);
        self._browse_button.setText("Browse...");
        QObject.connect(self._browse_button,SIGNAL("clicked()"),self._browse);

        #domain buttons menu
        self._axesbox = QVBox(self._mainvbox);
        self._axesbox.setSpacing(6);
        self._get_axis_list();
        self._axes =[];
        self.dim=[];
        self.xfrom=[];
        self.xto=[];
        self.nsteps=[];
        for axis in self._axis_list:
            self._add_axis(str(axis));
            
        #self._buttonbox = QHBox(self._mainvbox);
        #self._add_button = QPushButton("Add Axis",self._buttonbox);
        #QObject.connect(self._add_button, SIGNAL('clicked()'), self._add_axis)
        #self._remove_button = QPushButton("Remove Axis",self._buttonbox);
        #QObject.connect(self._remove_button, SIGNAL('clicked()'), self._remove_axis)
        #extra buttons menu
        self._buttonbox2 = QVBox(self._mainvbox);

        #parmnames menu
        self._parmbox= QVBox(self._mainvbox);
        self._parmbox.setSpacing(6);
        self._parm_label = QLabel("Parameters",self._parmbox);
        self.parmtable = QTable(self._parmbox,"parmlist");
        self.parmtable.setShowGrid(False);
        self.parmtable.setFocusStyle(QTable.FollowStyle);
        self.parmtable.setSelectionMode (QTable.SingleRow );
        self.parmtable.setNumCols(2);      
        self.parmtable.horizontalHeader () .setLabel(0,"name");
        self.parmtable.horizontalHeader () .setLabel(1,"nr. funklets");
        self.parmtable.setColumnWidth(0,200);
        self.parmtable.setColumnWidth(1,150);

        QObject.connect(self.parmtable,SIGNAL("selectionChanged()"),self._get_selected);


        self._parm_button = QPushButton(self._parmbox);
        self._parm_button.setText("Get Parms");

        self._parmlist = [];
        self._funklist = {};
        self._Comp_funklet={};
        QObject.connect(self._parm_button,SIGNAL("clicked()"),self._get_parms);

        self._domain = meq.domain(0,1,0,1);
        
        self._plot_button = QPushButton(self._parmbox);
        self._plot_button.setText("Plot Selection");

        QObject.connect(self._plot_button,SIGNAL("clicked()"),self._plot_parms);
       

        caption = "Inspect Table";
        self.set_widgets(self.wtop(),caption);
  

    def wtop(self):
        return self._wtop;

    def _get_table_name(self):
        name = str(self._name_edit.text());
        self._table_name=name;


    def _get_axis_list(self):

        forest_state=meqds.get_forest_state();
        axis_map=forest_state.axis_map;
        self._axis_list=[];
        for i in axis_map:
            if 'id' in i:
                if i.id:
                    self._axis_list.append(i.id);


    def _add_axis(self,axisnr=0):
        if isinstance(axisnr,int):
            axis_name=self._axis_list[axisnr];
        else:
            axis_name=axisnr;
        
        axis= QHBox(self._axesbox);
        axis.setSpacing(4);
        nr=0;
        nr=len(self._axes);
        if nr>= len(self._axis_list):
            #max number of axes reached
            return;
        self.dim.append(QLabel(axis_name,axis));
        textfrom = QLabel("from",axis);
        self.xfrom.append(QLineEdit(axis));

        textto = QLabel("to",axis);
        self.xto.append(QLineEdit(axis));
        textsteps = QLabel("steps",axis);
        self.nsteps.append(QLineEdit(axis));
        self._axes.append(axis)
        axis.show();


    def _get_range(self,pattern="*"):
        #gets the total domain for a given (sest of)parm(s) and fills the axes wiht these as default values
        self._domain = self._db.getRange(pattern);
        #TESTTESTTESTTEST
        #        self._domain['l'] = [-0.057,0.043];
        #        self._domain['m'] = [-0.038,0.062];


    def setDefaultNaxis(self):
        self._Naxis=[];
        for n in range(len(self._axes)):
             self._Naxis.append(10);


    def _get_selected(self,parm=None):
        if parm == None:
            row = self.parmtable.currentRow();
            parm = str(self.parmtable.item(row,0).text());
        self._get_range(parm);
        if parm not in self._funklist:
            self._get_funklets(parm=parm);
        if parm not in self._Comp_funklet:
            if(len(self._funklist[parm])>1):
                self._Comp_funklet[parm]=ComposedFunklet(self._funklist[parm]);
            else:
                self._Comp_funklet[parm]=self._funklist[parm][0];
                

        self._Comp_funklet[parm].setDomain(self._domain);
        self._Comp_funklet[parm].getNX();
        self._Naxis=self._Comp_funklet[parm].getNAxis();

        if not self._Naxis:
            self.setDefaultNaxis();
        #TESTTESTTESTTEST
#        self._Naxis[2]=10;
#        self._Naxis[3]=10;
        
        for n in range(len(self._axes)):
            axis_name = str(self._axis_list[n]).lower(); 
            if axis_name not in self._domain:
                break;
            self.xfrom[n].setText(str(self._domain[axis_name][0]));
            self.xto[n].setText(str(self._domain[axis_name][1]));
            self.nsteps[n].setText(str(self._Naxis[n]));
        return;
        

    def _remove_axis(self):
        if len(self._axes)<=1:
            return;
        nr=len(self._axes)-1;
        self._axes[nr].close();
        self._axes.pop();
        self.dim.pop();
        self.xfrom.pop();
        self.xto.pop();


    def _browse(self):
        fn = self.open_browse();
        if fn is None:
            return;
        self._table_name=fn;
        self._name_edit.setText(fn);
        self._get_parms();
    
    def open_browse(self):
        """Read a saved inarg record from a file, using a file browser"""
        filename = QFileDialog.getExistingDirectory("", self._wtop);
        if filename == QString.null:
            return None;
        filename = str(filename);
        return filename;


    def _get_parms(self):
        self._get_table_name();
        if not os.path.isdir(self._table_name):
            QMessageBox.warning(self.wtop(),
                                "Warning",
                                "Table not found, Please specify existing table");
            return;
        
        self._db = Parmdb(self._table_name);
        self._parmlist = self._db.getParmNames();
        self._funklist = {};
        #        if self._parmlist:
        self._update_parmtable();

    def _update_parmtable(self):
        if self._parmlist:
            self.parmtable.setNumRows(len(self._parmlist));
        else:
            self.parmtable.setNumRows(0);
            return;
        parmnr=0;
        for parm in self._parmlist:
            self.parmtable.setText(parmnr,0,parm);
            if parm in self._funklist:
                nr_funklets = len(self._funklist[parm]);
                self.parmtable.setText(parmnr,1,str(nr_funklets));
            else:
                self.parmtable.clearCell(parmnr,1);
            parmnr+=1;
        self._get_range();

    def _get_all_funklets(self):
        self._funklist = {};
        for parm in self._parmlist:
            self._funklist[parm] = self._db.getFunklets(parm);
            self._helplist=[];
            for funk in self._funklist[parm]:
                # convert  to evaluatable Funklet:
                self._helplist.append(Funklet(funklet = funk,name=parm));
            self._funklist[parm] = self._helplist;
        #print "all funklets",self._funklist;

    def _get_funklets(self,parm="all"):
        if parm == "all":
            self._get_all_funklets();
        else:
            #print "getting funklets",parm;
            self._funklist[parm] = self._db.getFunklets(parm);
            self._helplist=[];
            for funk in self._funklist[parm]:
                # convert  to evaluatable Funklet:
                self._helplist.append(Funklet(funklet = funk,name=parm));
                self._funklist[parm] = self._helplist;
           
        self._update_parmtable();        
    
    def _plot_parms(self,parm=None):
        if parm == None:
            row = self.parmtable.currentRow();
            parm = str(self.parmtable.item(row,0).text());
        #print "plotting",parm
        if parm is None:
            return;
        if parm not in self._funklist:
            self._get_funklets(parm);
            #print "plotting",self._funklist[parm];
        
            self._Comp_funklet[parm] =  ComposedFunklet(self._funklist[parm]);
        #print self._Comp_funklet[parm].eval({'x0':.4,'x1':.2,'x2':3,'x5':4});
        cells = self._getCells();

        #print cells;
        self._Comp_funklet[parm].plot(cells=cells,parent=self._wtop);
        

    def _getCells(self):
      dom = self._domain;
      cells = meq.cells(dom,num_time=self._Naxis[0],num_freq=self._Naxis[1]);
      #add more axes here...
      #print "domain",dom
      for dim in range(2,len(self._axis_list)):
          id = str(self._axis_list[dim]).lower();
          if id in dom:
              step_size=float(dom[id][1]-dom[id][0])/self._Naxis[dim];
              #print "steP", step_size,self._Naxis[dim],dom;
              startgrid=0.5*step_size+dom[id][0];
              grid = [];
              cell_size=[];
              for i in range(self._Naxis[dim]):
                  grid.append(i*step_size+startgrid);
                  cell_size.append(step_size);
              cells.cell_size[id]=array(cell_size);
              cells.grid[id]=array(grid);
              cells.segments[id]=record(start_index=0,end_index=self._Naxis[dim]-1);

      cells.domain=dom;

      
      return cells;


def _start_table_inspector ():
  _dprint(2,'adding a table inspector');
  # Create a 'data item', which represents our box in the grid.
  # The first argument is a UDI, we only need to ensure it is unique
  # so that it doesn't get confused with other data items in the system
  item = Grid.DataItem('/TableInspector',                    # UDI
                        name='Table Inspector',         # name
                        caption='<b>Table Inspector</b>',       # caption for GUI -- note use of Rich Text markup
                        desc='Table inspector plug-in',
                        data=None,refresh=None,             # no data and no refresh function
                        viewer=TableInspector);           # viewer class
  # this registers us with the grid and creates a viewer
  Grid.addDataItem(item);



# this is called automatically by the browser to populate the tree browser
# context menus and toolbar
def define_treebrowser_actions (tb):
  #return;  # temporary while this plugin is broken
  _dprint(1,'defining table inspector treebrowser actions');
  parent = tb.wtop();
  # create QAction for the Stream control plugin
  global _qa_table;
  _qa_table = QAction("Table Inspector",pixmaps.table.iconset(),"Table Inspector",0,parent);
  _qa_table.setMenuText("Table Inspector");
  # make sure it's enabled/disabled as appropriate
  _qa_table._is_enabled = lambda tb=tb: tb.is_connected;
  # "45" is priority of action, determining its place in the toolbar
  tb.add_action(_qa_table,45,callback=_start_table_inspector);





def define_mainmenu_actions (menu,parent):
  #return;  # temporary while this plugin is broken
  _dprint(1,'defining table inspector menu actions');
  global _qa_table;
  # add ourselves to the MeqTrees menu
  _qa_table.addTo(menu['MeqTrees']);




#!/usr/bin/env python3


#% $Id: printfilter.py 4329 2006-12-14 17:27:23Z twillis $ 

#
# Copyright (C) 2002-2007
# ASTRON (Netherlands Foundation for Research in Astronomy)
# and The MeqTree Foundation
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands, seg@astron.nl
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

# this code is taken directly from the PyQwt BodeDemo.py example
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from qwt.qt.QtCore import Qt,  QSize
from qwt.qt.QtGui import QColor,  QPixmap, QFont,  QIcon
from qwt.qt.QtGui import QMainWindow,  QWidget,  QToolBar,  QToolButton,  QHBoxLayout,  QLabel,  QApplication
from qwt.qt.QtGui import QPrintDialog, QPrinter



class PrintFilter(Qwt.QwtPlotPrintFilter):
    def __init__(self):
        QwtPlotPrintFilter.__init__(self)

    # __init___()

    def color(self, c, item):
        if not (self.options() & QwtPlotPrintFilter.CanvasBackground):
            if item == QwtPlotPrintFilter.MajorGrid:
                return Qt.darkGray
            elif item == QwtPlotPrintFilter.MinorGrid:
                return Qt.gray
        if item == QwtPlotPrintFilter.Title:
            return Qt.red
        elif item == QwtPlotPrintFilter.AxisScale:
            return Qt.green
        elif item == QwtPlotPrintFilter.AxisTitle:
            return Qt.blue
        return c

    # color()

    def font(self, f, _):
        result = QFont(f)
        result.setPointSize(int(f.pointSize()*1.25))
        return result

    # font()

# class PrintFilter

